import pytest

from xdsl.utils.runtime_final import is_runtime_final, runtime_final


class NotFinal:
    """A non-Final class."""


@runtime_final
class Final:
    """A Final class."""


def test_is_runtime_final():
    """Check that `is_runtime_final` returns the correct value."""
    assert not is_runtime_final(NotFinal)
    assert is_runtime_final(Final)


def test_final_inheritance_error():
    """Check that final classes cannot be subclassed."""
    with pytest.raises(TypeError, match="Subclassing final classes is restricted"):

        class SubFinal(Final):
            pass

        SubFinal()



from io import StringIO

import pytest

from xdsl.backend.csl import print_csl
from xdsl.dialects import builtin as bi
from xdsl.dialects import csl, test
from xdsl.ir import Attribute, Block, Operation, Region


@pytest.fixture
def layout():
    return csl.CslModuleOp(
        regions=[Region(Block([]))],
        properties={"kind": csl.ModuleKindAttr(csl.ModuleKind.LAYOUT)},
        attributes={"sym_name": bi.StringAttr("layout")},
    )


@pytest.fixture
def program():
    return csl.CslModuleOp(
        regions=[Region(Block([]))],
        properties={"kind": csl.ModuleKindAttr(csl.ModuleKind.PROGRAM)},
        attributes={"sym_name": bi.StringAttr("program")},
    )


class NewPrinter(print_csl.CslPrintContext):
    def mlir_type_to_csl_type(self, type_attr: Attribute) -> str:
        match type_attr:
            case test.TestType(data=data):
                return data
            case _:
                return super().mlir_type_to_csl_type(type_attr)

    def print_op(self, op: Operation):
        match op:
            case test.TestOp(res=res, ops=ops):
                var = "test" if not ops else self._var_use(ops[0])
                self._print_or_promote_to_inline_expr(res[0], var)
            case _:
                super().print_op(op)


def test_csl_printer_extension(layout: csl.CslModuleOp, program: csl.CslModuleOp):
    io = StringIO()
    io.write("\n")
    program.body.block.add_op(first := test.TestOp(result_types=[bi.i32]))
    program.body.block.add_op(
        test.TestOp(operands=[first.results[0]], result_types=[test.TestType("Test")])
    )
    mod = bi.ModuleOp([program, layout])
    print_csl.print_to_csl(mod, io, NewPrinter)
    expected = r"""
// FILE: program
const v0 : i32 = test;
const v1 : Test = v0;
// -----
// FILE: layout
"""
    assert io.getvalue() == expected



from xdsl.context import Context
from xdsl.dialects.builtin import ModuleOp, StringAttr
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.transforms.dead_code_elimination import dce

from ..dialects import toy


class InlineFunctions(RewritePattern):
    _func_op_by_name: dict[str, toy.FuncOp] | None = None

    def lookup_func_op(self, module: ModuleOp, name: str) -> toy.FuncOp:
        if self._func_op_by_name is None:
            self._func_op_by_name = {
                op.sym_name.data: op for op in module.ops if isinstance(op, toy.FuncOp)
            }
        return self._func_op_by_name[name]

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: toy.GenericCallOp, rewriter: PatternRewriter):
        """
        For each generic call, find the function that it calls, and inline it.
        """

        # Get module
        parent = op.parent_op()
        assert isinstance(parent, toy.FuncOp)
        module = parent.parent_op()
        assert isinstance(module, ModuleOp)

        # Clone called function
        impl = self.lookup_func_op(module, op.callee.string_value()).clone()
        impl_block = impl.body.block

        # Cast operands to unranked
        inputs = [toy.CastOp(operand) for operand in op.operands]

        # Insert casts before matched op
        rewriter.insert_op_before_matched_op(inputs)

        # Replace block args with operand casts
        for i, arg in zip(inputs, impl_block.args):
            arg.replace_by(i.res)

        # remove block args
        while len(impl_block.args):
            assert not impl_block.args[-1].uses
            rewriter.erase_block_argument(impl_block.args[-1])

        # Inline function definition before matched op
        rewriter.inline_block_before_matched_op(impl_block)

        # Get return from function definition
        return_op = op.prev_op
        assert return_op is not None

        rewriter.replace_matched_op([], return_op.operands)
        rewriter.erase_op(return_op)


class RemoveUnusedPrivateFunctions(RewritePattern):
    _used_funcs: set[str] | None = None

    def should_remove_op(self, op: toy.FuncOp) -> bool:
        if op.sym_visibility != StringAttr("private"):
            return False

        if self._used_funcs is None:
            # Get module
            module = op.parent_op()
            assert isinstance(module, ModuleOp)

            self._used_funcs = {
                op.callee.string_value()
                for op in module.walk()
                if isinstance(op, toy.GenericCallOp)
            }

        return op.sym_name.data not in self._used_funcs

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: toy.FuncOp, rewriter: PatternRewriter):
        if self.should_remove_op(op):
            rewriter.erase_matched_op()


class InlineToyPass(ModulePass):
    """
    A custom pass to inline Toy functions. In MLIR, this is done through an interface, as
    described in the Toy tutorial. As of time of writing, we don't have dialect interfaces
    in xDSL, but would like to add them. This pass should be migrated to the built-in pass
    when they land.

    https://mlir.llvm.org/docs/Tutorials/Toy/Ch-4/
    https://github.com/xdslproject/xdsl/issues/957

    """

    name = "inline-toy-functions"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        PatternRewriteWalker(InlineFunctions()).rewrite_module(op)
        PatternRewriteWalker(RemoveUnusedPrivateFunctions()).rewrite_module(op)
        dce(op)



from xdsl.backend.register_allocatable import RegisterAllocatableOperation
from xdsl.backend.register_allocator import BlockAllocator
from xdsl.ir import Block
from xdsl.utils.exceptions import DiagnosticException


class BlockNaiveAllocator(BlockAllocator):
    """
    It traverses the use-def SSA chain backwards (i.e., from uses to defs) and:
      1. allocates registers for operands
      2. frees registers for results (since that will be the last time they appear when
      going backwards)
    It currently operates on blocks.

    This is a simplified version of the standard bottom-up local register allocation
    algorithm.

    A relevant reference in "Engineering a Compiler, Edition 3" ISBN: 9780128154120.

    ```
    for op in block.walk_reverse():
    for operand in op.operands:
        if operand is not allocated:
            allocate(operand)

    for result in op.results:
    if result is not allocated:
        allocate(result)
        free_before_next_instruction.append(result)
    else:
        free(result)
    ```
    """

    def allocate_block(self, block: Block):
        for op in reversed(block.ops):
            if isinstance(op, RegisterAllocatableOperation):
                try:
                    op.allocate_registers(self)
                except DiagnosticException as e:
                    op.emit_error("Error allocating op", e)



import pytest

from xdsl.dialect_interfaces import DialectInterface, OpAsmDialectInterface
from xdsl.ir import Dialect


class TestInterface(DialectInterface):
    pass


def test_interfaces():
    dialect = Dialect("dialect", [], [], [TestInterface()])
    another_dialect = Dialect("another", [], [], [])

    assert dialect.has_interface(TestInterface)
    assert not another_dialect.has_interface(TestInterface)
    assert isinstance(dialect.get_interface(TestInterface), TestInterface)


def test_op_asm_interface():
    interf = OpAsmDialectInterface()

    key = interf.declare_resource("some_key")
    assert key == "some_key"
    interf.parse_resource(key, "0x0800000001")
    assert interf.lookup(key) == "0x0800000001"

    key = interf.declare_resource("some_key")
    assert key == "some_key_0"
    interf.parse_resource(key, "0x0800000002")
    assert interf.lookup(key) == "0x0800000002"

    with pytest.raises(
        ValueError, match="Blob must be a hex string, got: normal string"
    ):
        interf.parse_resource("some_key", "normal string")

    with pytest.raises(KeyError):
        interf.parse_resource("non existent key", "0x0800000003")

    interf.declare_resource("non_assigned_key")
    assert interf.build_resources(["some_key", "non_assigned_key"]) == {
        "some_key": "0x0800000001"
    }



from xdsl.dialects import memref_stream
from xdsl.dialects.builtin import AffineMapAttr, ArrayAttr
from xdsl.ir import SSAValue
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)


class RemoveUnusedInitOperandPattern(RewritePattern):
    """
    Removes the inputs corresponding to unused arguments in the body.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: memref_stream.GenericOp, rewriter: PatternRewriter
    ) -> None:
        if memref_stream.IteratorTypeAttr.interleaved() in op.iterator_types.data:
            # Do not run on interleaved ops
            return

        block = op.body.block
        block_args = block.args

        inputs = op.inputs

        num_inputs = len(inputs)

        unused_input_indices = tuple(
            index for index, arg in enumerate(block_args[:num_inputs]) if not arg.uses
        )

        if not unused_input_indices:
            # All args have uses, nothing to remove
            return

        outputs = op.outputs

        optional_inits: list[SSAValue | None] = [None] * len(outputs)
        for init_index, init in zip(op.init_indices, op.inits, strict=True):
            optional_inits[init_index.data] = init

        assert len(optional_inits) == len(outputs)

        new_inputs: list[SSAValue] = []
        new_indexing_maps: list[AffineMapAttr] = []

        for index, arg in enumerate(block_args[:num_inputs]):
            drop_operand = index in unused_input_indices
            if drop_operand:
                arg.erase()
                continue

            new_indexing_maps.append(op.indexing_maps.data[index])
            new_inputs.append(inputs[index])

        new_indexing_maps.extend(op.indexing_maps.data[num_inputs:])

        for i in reversed(unused_input_indices):
            block.erase_arg(block_args[i])

        rewriter.replace_matched_op(
            memref_stream.GenericOp(
                new_inputs,
                op.outputs,
                op.inits,
                rewriter.move_region_contents_to_new_regions(op.body),
                ArrayAttr(new_indexing_maps),
                op.iterator_types,
                op.bounds,
                op.init_indices,
                op.doc,
                op.library_call,
            )
        )



from dataclasses import dataclass
from typing import cast

from xdsl.context import Context
from xdsl.dialects import builtin, vector, x86
from xdsl.dialects.builtin import (
    FixedBitwidthType,
    UnrealizedConversionCastOp,
    VectorType,
)
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.exceptions import DiagnosticException

from .helpers import vector_type_to_register_type


@dataclass
class VectorBroadcastToX86(RewritePattern):
    arch: str

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: vector.BroadcastOp, rewriter: PatternRewriter):
        # Get the register to be broadcasted
        source_cast_op, source_x86 = UnrealizedConversionCastOp.cast_one(
            op.source, x86.register.UNALLOCATED_GENERAL
        )
        # Actually broadcast the register
        element_type = op.source.type
        assert isinstance(element_type, FixedBitwidthType)
        element_size = element_type.bitwidth
        match element_size:
            case 16:
                raise DiagnosticException(
                    "Half-precision vector broadcast is not implemented yet."
                )
            case 32:
                broadcast = x86.ops.DS_VpbroadcastdOp
            case 64:
                broadcast = x86.ops.DS_VpbroadcastqOp
            case _:
                raise DiagnosticException(
                    "Float precision must be half, single or double."
                )
        broadcast_op = broadcast(
            source=source_x86,
            destination=vector_type_to_register_type(op.vector.type, self.arch),
        )
        # Get back the abstract vector
        dest_cast_op, _ = UnrealizedConversionCastOp.cast_one(
            broadcast_op.destination, op.vector.type
        )

        rewriter.replace_matched_op([source_cast_op, broadcast_op, dest_cast_op])


@dataclass
class VectorFMAToX86(RewritePattern):
    arch: str

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: vector.FMAOp, rewriter: PatternRewriter):
        vect_type = cast(VectorType, op.acc.type)
        x86_vect_type = vector_type_to_register_type(vect_type, self.arch)
        # Pointer casts
        lhs_cast_op, lhs_new = UnrealizedConversionCastOp.cast_one(
            op.lhs, x86_vect_type
        )
        rhs_cast_op, rhs_new = UnrealizedConversionCastOp.cast_one(
            op.rhs, x86_vect_type
        )
        acc_cast_op, acc_new = UnrealizedConversionCastOp.cast_one(
            op.acc, x86_vect_type
        )
        # Instruction selection
        element_size = cast(FixedBitwidthType, vect_type.get_element_type()).bitwidth
        match element_size:
            case 16:
                raise DiagnosticException(
                    "Half-precision vector load is not implemented yet."
                )
            case 32:
                fma = x86.ops.RSS_Vfmadd231psOp
            case 64:
                fma = x86.ops.RSS_Vfmadd231pdOp
            case _:
                raise DiagnosticException(
                    "Float precision must be half, single or double."
                )
        fma_op = fma(acc_new, lhs_new, rhs_new)

        res_cast_op = UnrealizedConversionCastOp.get(
            (fma_op.register_out,), (vect_type,)
        )
        rewriter.replace_matched_op(
            [lhs_cast_op, rhs_cast_op, acc_cast_op, fma_op, res_cast_op]
        )


@dataclass(frozen=True)
class ConvertVectorToX86Pass(ModulePass):
    name = "convert-vector-to-x86"

    arch: str

    def apply(self, ctx: Context, op: builtin.ModuleOp) -> None:
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    VectorFMAToX86(self.arch),
                    VectorBroadcastToX86(self.arch),
                ]
            ),
            apply_recursively=False,
        ).rewrite_module(op)



from __future__ import annotations

from collections.abc import Iterable, Sequence
from dataclasses import dataclass, field

from xdsl.ir import (
    Attribute,
    Block,
    BlockArgument,
    Operation,
    OpResult,
    Region,
    SSAValue,
    SSAValues,
)


@dataclass(frozen=True)
class InsertPoint:
    """
    An insert point.
    It is either a point before an operation, or at the end of a block.

    See external [documentation](https://mlir.llvm.org/doxygen/classmlir_1_1OpBuilder_1_1InsertPoint.html).
    """

    block: Block
    """The block where the insertion point is in."""

    insert_before: Operation | None = field(default=None)
    """
    The insertion point is right before this operation.
    If the operation is None, the insertion point is at the end of the block.
    """

    def __post_init__(self) -> None:
        # Check that the insertion point is valid.
        # An insertion point can only be invalid if `insert_before` is an `Operation`,
        # and its parent is not `block`.
        if self.insert_before is not None:
            if self.insert_before.parent is not self.block:
                raise ValueError("Insertion point must be in the builder's `block`")

    @staticmethod
    def before(op: Operation) -> InsertPoint:
        """Gets the insertion point before an operation."""
        if (block := op.parent_block()) is None:
            raise ValueError("Operation insertion point must have a parent block")
        return InsertPoint(block, op)

    @staticmethod
    def after(op: Operation) -> InsertPoint:
        """Gets the insertion point after an operation."""
        block = op.parent_block()
        if block is None:
            raise ValueError("Operation insertion point must have a parent block")
        return InsertPoint(block, op.next_op)

    @staticmethod
    def at_start(block: Block) -> InsertPoint:
        """Gets the insertion point at the start of a block."""
        return InsertPoint(block, block.ops.first)

    @staticmethod
    def at_end(block: Block) -> InsertPoint:
        """Gets the insertion point at the end of a block."""
        return InsertPoint(block)


@dataclass(frozen=True)
class BlockInsertPoint:
    """
    An insert point for a block.
    It is either a point before a block, or after a block.
    """

    region: Region
    """The region where the insertion point is in."""

    insert_before: Block | None = field(default=None)
    """
    The insertion point is right before this block.
    If the block is None, the insertion point is at the end of the region.
    """

    def __post_init__(self) -> None:
        # Check that the insertion point is valid.
        # An insertion point can only be invalid if `insert_before` is a `Block`,
        # and its parent is not `region`.
        if self.insert_before is not None:
            if self.insert_before.parent is not self.region:
                raise ValueError("Insertion point must be in the builder's `region`")

    @staticmethod
    def before(block: Block) -> BlockInsertPoint:
        """Gets the insertion point before a block."""
        if (region := block.parent) is None:
            raise ValueError("Block insertion point must have a parent region")
        return BlockInsertPoint(region, block)

    @staticmethod
    def after(block: Block) -> BlockInsertPoint:
        """Gets the insertion point after a block."""
        region = block.parent
        if region is None:
            raise ValueError("Block insertion point must have a parent region")
        return BlockInsertPoint(region, block.next_block)

    @staticmethod
    def at_start(region: Region) -> BlockInsertPoint:
        """Gets the insertion point at the start of a region."""
        return BlockInsertPoint(region, region.first_block)

    @staticmethod
    def at_end(region: Region) -> BlockInsertPoint:
        """Gets the insertion point at the end of a region."""
        return BlockInsertPoint(region)


class Rewriter:
    @staticmethod
    def erase_op(op: Operation, safe_erase: bool = True):
        """
        Erase an operation.
        Check that the operation has no uses, and has a parent.
        If safe_erase is True, check that the operation has no uses.
        Otherwise, replace its uses with ErasedSSAValue.
        """
        if (block := op.parent) is not None:
            block.erase_op(op, safe_erase=safe_erase)
        else:
            op.erase(safe_erase=safe_erase)

    @staticmethod
    def replace_op(
        op: Operation,
        new_ops: Operation | Sequence[Operation],
        new_results: Sequence[SSAValue | None] | None = None,  # noqa
        safe_erase: bool = True,
    ):
        """
        Replace an operation with multiple new ones.
        If new_results is specified, map the results of the deleted operations with these
        SSA values.
        Otherwise, use the results of the last operation added.
        None elements in new_results are the SSA values to delete.
        If safe_erase is False, then operations can be deleted even if they are
        still used.
        """
        if op.parent is None:
            raise ValueError("Cannot replace an operation without a parent")
        block = op.parent

        if isinstance(new_ops, Operation):
            new_ops = [new_ops]
        if new_results is None:
            new_results = [] if len(new_ops) == 0 else new_ops[-1].results

        if len(op.results) != len(new_results):
            raise ValueError(
                f"Expected {len(op.results)} new results, but got {len(new_results)}"
            )

        for old_result, new_result in zip(op.results, new_results):
            if new_result is None:
                old_result.erase(safe_erase=safe_erase)
            else:
                old_result.replace_by(new_result)

        block.insert_ops_after(new_ops, op)

        if len(op.results):
            for new_op in new_ops:
                for res in new_op.results:
                    res.name_hint = op.results[0].name_hint

        block.erase_op(op, safe_erase=safe_erase)

    @staticmethod
    def replace_value_with_new_type(val: SSAValue, new_type: Attribute) -> SSAValue:
        """
        Replace a value with a value of a new type, and return the new value.
        This will insert the new value in the operation or block, and remove the existing
        value.
        """
        if isinstance(val, OpResult):
            operation = val.op
            index = val.index
            new_value = OpResult(new_type, operation, val.index)
            results = operation.results
            operation.results = SSAValues(
                (*results[:index], new_value, *results[index + 1 :])
            )
        elif isinstance(val, BlockArgument):
            block = val.block
            index = val.index
            new_value = BlockArgument(new_type, block, index)
            args = block.args
            block._args = (  # pyright: ignore[reportPrivateUsage]
                *args[:index],
                new_value,
                *args[index + 1 :],
            )
        else:
            raise ValueError(
                f"Expected OpResult or BlockArgument, got {type(val).__name__}"
            )

        new_value.name_hint = val.name_hint
        val.replace_by(new_value)
        return new_value

    @staticmethod
    def inline_block(
        source: Block, insertion_point: InsertPoint, arg_values: Sequence[SSAValue] = ()
    ):
        """
        Move the block operations before another operation.
        The block should not be a parent of the operation.
        """
        # MLIR equivalent:
        # https://github.com/llvm/llvm-project/blob/96a3d05ed923d2abd51acb52984b83b9e8044924/mlir/lib/IR/PatternMatch.cpp#L290
        assert arg_values == () or len(arg_values) == len(source.args), (
            f"Expected {len(source.args)} replacement argument values, got "
            f"{len(arg_values)}"
        )

        # The source block will be deleted, so it should not have any users (i.e.,
        # there should be no predecessors).
        # TODO: check that the block has no predecessors

        #  assert not block.predecessors, "expected 'source' to have no predecessors"

        dest = insertion_point.block

        # TODO: verify that the successors will make sense after inlining
        # We currently cannot perform this check, just like the TODO above, due to lack
        # of infrastructure in xDSL
        # https://github.com/xdslproject/xdsl/issues/2066

        # if dest.last_op != op:
        #       The source block will be inserted in the middle of the dest block, so the
        #       source block should have no successors. Otherwise, the remainder of the dest
        #       block would be unreachable.
        #       assert not source.successors, "expected 'source' to have no successors");
        # else:
        #       The source block will be inserted at the end of the dest block, so the dest
        #       block should have no successors. Otherwise, the inserted operations will be
        #       unreachable.
        #       assert not dest.successors,  "expected 'dest' to have no successors");

        # Replace all of the successor arguments with the provided values.
        if arg_values:
            for arg, val in zip(source.args, arg_values, strict=True):
                arg.replace_by(val)

        # Move operations from the source block to the dest block and erase the
        # source block.
        ops = list(source.ops)
        for block_op in ops:
            block_op.detach()

        if (insert_before := insertion_point.insert_before) is not None:
            dest.insert_ops_before(ops, insert_before)
        else:
            dest.add_ops(ops)

        parent_region = source.parent
        if parent_region is not None:
            parent_region.detach_block(source)
        source.erase()

    @staticmethod
    def insert_block(block: Block | Iterable[Block], insert_point: BlockInsertPoint):
        """
        Insert one or multiple blocks at a given location.
        The blocks to insert should be detached from any region.
        The insertion point should not be contained in the block to insert.
        """
        region = insert_point.region
        if insert_point.insert_before is not None:
            region.insert_block_before(block, insert_point.insert_before)
        else:
            region.add_block(block)

    @staticmethod
    def insert_op(
        op_or_ops: Operation | Sequence[Operation], insertion_point: InsertPoint
    ):
        """Insert operations at a certain location in a block."""
        ops = (op_or_ops,) if isinstance(op_or_ops, Operation) else op_or_ops
        if insertion_point.insert_before is not None:
            insertion_point.block.insert_ops_before(ops, insertion_point.insert_before)
        else:
            insertion_point.block.add_ops(ops)

    @staticmethod
    def move_region_contents_to_new_regions(region: Region) -> Region:
        """Move the region blocks to a new region."""
        new_region = Region()
        region.move_blocks(new_region)
        return new_region

    @staticmethod
    def inline_region(region: Region, insertion_point: BlockInsertPoint) -> None:
        """Move the region blocks to a given location."""
        if insertion_point.insert_before is not None:
            region.move_blocks_before(insertion_point.insert_before)
        else:
            region.move_blocks(insertion_point.region)



from xdsl.dialects import memref
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.utils.hints import isa


class MemRefSubviewOfSubviewFolding(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: memref.SubviewOp, rewriter: PatternRewriter, /):
        source_subview = op.source.owner
        if not isinstance(source_subview, memref.SubviewOp):
            return

        current_strides = op.static_strides.get_values()

        if not all(stride == 1 for stride in current_strides):
            return

        if not all(
            stride == 1 for stride in source_subview.static_strides.iter_values()
        ):
            return

        if not len(op.static_offsets) == len(source_subview.static_offsets):
            return

        assert isa(source_subview.source.type, memref.MemRefType)

        assert isa(op.result.type, memref.MemRefType)

        reduce_rank = False

        if len(source_subview.source.type.shape) != len(op.result.type.shape):
            reduce_rank = True

        if len(op.offsets) > 0 or len(source_subview.offsets) > 0:
            return
        if len(op.sizes) > 0 or len(source_subview.sizes) > 0:
            return
        if len(op.strides) > 0 or len(source_subview.strides) > 0:
            return

        new_offsets = [
            off1 + off2
            for off1, off2 in zip(
                op.static_offsets.iter_values(),
                source_subview.static_offsets.iter_values(),
                strict=True,
            )
        ]

        current_sizes = op.static_sizes.get_values()

        new_op = memref.SubviewOp.from_static_parameters(
            source_subview.source,
            source_subview.source.type,
            new_offsets,
            current_sizes,
            current_strides,
            reduce_rank=reduce_rank,
        )
        if reduce_rank:
            if new_op.result.type != op.result.type:
                return

        rewriter.replace_matched_op(new_op)


class ElideUnusedAlloc(RewritePattern):
    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: memref.AllocOp, rewriter: PatternRewriter, /):
        if isinstance(only_use := op.memref.get_user_of_unique_use(), memref.DeallocOp):
            rewriter.erase_op(only_use)
            rewriter.erase_matched_op()



from dataclasses import dataclass

from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import arith, scf
from xdsl.dialects.builtin import (
    FunctionType,
    IndexType,
    IntegerAttr,
    MemRefType,
    ModuleOp,
    SymbolRefAttr,
    i32,
)
from xdsl.dialects.csl import csl, csl_stencil, csl_wrapper
from xdsl.ir import (
    Block,
    Operation,
    OpResult,
    Region,
)
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)
from xdsl.rewriter import InsertPoint
from xdsl.utils.exceptions import PassFailedException
from xdsl.utils.hints import isa


@dataclass()
class HandleCslStencilApplyAsyncCF(RewritePattern):
    """
    Ensure that the async csl_stencil.apply op is last in its function.
    Any code following an apply is split off into a separate function.
    Control flow proceeds from the apply's second callback to the split-off function.
    """

    counter: int

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: csl_stencil.ApplyOp, rewriter: PatternRewriter, /):
        # invalid case
        if not op.next_op:
            return

        # case 1: apply is last in function, nothing to do
        if isinstance(op.next_op, csl.ReturnOp):
            return

        terminator = op.done_exchange.block.last_op
        assert isinstance(terminator, csl_stencil.YieldOp)

        # case 2: apply is followed by call_op and return - move call_op to second callback of apply
        if (
            isinstance(call_op := op.next_op, csl.CallOp)
            and op.next_op.next_op
            and isinstance(op.next_op.next_op, csl.ReturnOp)
        ):
            rewriter.insert_op(call_op.clone(), InsertPoint.before(terminator))
            rewriter.erase_op(call_op)
            return

        parent_func = op.parent_op()
        while parent_func is not None and not isinstance(parent_func, csl.FuncOp):
            parent_func = parent_func.parent_op()
        if not parent_func:
            return

        # case 3: apply is followed by other code, split it off into a different func, call it from second callback of apply
        assert (parent_block := op.parent_block()) is not None
        next_block = parent_block.split_before(op.next_op)
        rewriter.insert_op(csl.ReturnOp(), InsertPoint.after(op))
        next_func = csl.FuncOp(f"step{self.counter}", FunctionType.from_lists([], []))
        self.counter += 1
        rewriter.inline_block(next_block, InsertPoint.at_start(next_func.body.block))
        rewriter.insert_op(
            csl.CallOp(SymbolRefAttr(next_func.sym_name)),
            InsertPoint.before(terminator),
        )
        rewriter.insert_op(next_func, InsertPoint.after(parent_func))


@dataclass()
class ConvertForLoopToCallGraphPass(RewritePattern):
    """
    Translates for-loop to csl functions, to handle async communications in the loop body.
    Splits the body of the enclosing function into: pre_block, loop, post_block.
    Further splits `loop` into for_cond, for_body, for_inc.
    These can in theory all be functions, yet we setup `for_cond` as a local task
    to avoid potentially large in-memory call stacks for long-running loops.
    Loop var and iter_args are promoted to module-level csl vars.

    Limitations:
      * Loop can only yield its input iter_args (in any order)
      * Loop bounds and step must be arith.constant values
      * Iter args must currently be stencil buffers (memrefs) of the same data type
    """

    counter: int

    task_ids: list[int]

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: scf.ForOp, rewriter: PatternRewriter, /):
        if not self._is_inside_wrapper_outside_apply(op):
            return

        parent_func = op.parent_op()
        while parent_func is not None and not isinstance(parent_func, csl.FuncOp):
            parent_func = parent_func.parent_op()
        if not parent_func:
            return

        # limitation: can yield iter_args in any order, but they cannot be modified in the loop body
        terminator = op.body.block.last_op
        assert isinstance(terminator, scf.YieldOp)
        assert all(arg in op.body.block.args for arg in terminator.arguments), (
            "Can only yield unmodified iter_args (in any order)"
        )

        # limitation: currently only loops built from arith.constant are supported
        assert isinstance(op.lb, OpResult)
        assert isinstance(op.ub, OpResult)
        assert isinstance(op.step, OpResult)
        assert isinstance(op.lb.op, arith.ConstantOp)
        assert isinstance(op.ub.op, arith.ConstantOp)
        assert isinstance(op.step.op, arith.ConstantOp)
        assert isa(op.lb.op.value, IntegerAttr[IndexType])
        assert isa(op.ub.op.value, IntegerAttr[IndexType])
        assert isa(op.step.op.value, IntegerAttr[IndexType])

        # limitation: all iter_args must be memrefs (stencil buffers) and have the same data type
        if op.iter_args:
            assert isa(op.iter_args[0].type, MemRefType)
            element_type = op.iter_args[0].type.get_element_type()
            assert all(
                isa(a.type, MemRefType) and element_type == a.type.get_element_type()
                for a in op.iter_args
            )

        no_params = FunctionType.from_lists([], [])
        if self.task_ids:
            cond_task_id = self.task_ids.pop(0)
        else:
            raise PassFailedException(
                "Insufficient number of task IDs supplied, please provide further IDs to be used."
            )

        pre_block = op.parent_block()
        if pre_block is None:
            return
        post_block = pre_block.split_before(op)

        post_func = csl.FuncOp(f"for_post{self.counter}", no_params)
        rewriter.inline_block(post_block, InsertPoint.at_start(post_func.body.block))

        cond_func = csl.TaskOp(
            f"for_cond{self.counter}",
            no_params,
            task_kind=csl.TaskKind.LOCAL,
            id=cond_task_id,
        )
        body_func = csl.FuncOp(f"for_body{self.counter}", no_params)
        inc_func = csl.FuncOp(f"for_inc{self.counter}", no_params)

        # create csl.vars for loop var and iter_args outside the parent func
        rewriter.insert_op(
            iv := csl.VariableOp.from_value(IntegerAttr(op.lb.op.value.value, i32)),
            InsertPoint.before(parent_func),
        )
        iter_vars = [csl.VariableOp.from_type(arg_t) for arg_t in op.iter_args.types]
        rewriter.insert_op(iter_vars, InsertPoint.before(parent_func))

        iv.res.name_hint = "iteration"
        for i, v in enumerate(iter_vars):
            v.res.name_hint = f"var{i}"

        # parent func (pre loop): setup iter vars and activate cond_func
        with ImplicitBuilder(pre_block):
            for dst, src in zip(iter_vars, op.iter_args):
                csl.StoreVarOp(dst, src)
            csl.ActivateOp(cond_task_id, csl.TaskKind.LOCAL)
            csl.ReturnOp()

        # for-loop condition func
        with ImplicitBuilder(cond_func.body.block):
            ub = arith.ConstantOp.from_int_and_width(op.ub.op.value.value, i32)
            iv_load = csl.LoadVarOp(iv)
            iv_load.res.name_hint = f"{iv.res.name_hint}_cond"
            cond = arith.CmpiOp(iv_load, ub, "slt")
            branch = scf.IfOp(cond, [], Region(Block()), Region(Block()))
            with ImplicitBuilder(branch.true_region):
                csl.CallOp(SymbolRefAttr(body_func.sym_name))
                scf.YieldOp()
            with ImplicitBuilder(branch.false_region):
                csl.CallOp(SymbolRefAttr(post_func.sym_name))
                scf.YieldOp()
            csl.ReturnOp()

        # for-loop inc func
        with ImplicitBuilder(inc_func.body.block):
            step = arith.ConstantOp.from_int_and_width(op.step.op.value.value, i32)
            iv_load = csl.LoadVarOp(iv)
            iv_load.res.name_hint = f"{iv.res.name_hint}_inc"
            stepped = arith.AddiOp(iv_load, step)
            csl.StoreVarOp(iv, stepped)

            # pre-load iter_vars and store them in the order specified in scf.yield
            load_vars = [csl.LoadVarOp(v) for v in iter_vars]
            for v in load_vars:
                v.res.name_hint = f"{v.var.name_hint}_inc"

            # for out-of-order yields, store yielded var to iter_var
            for iter_var, yielded_var in zip(iter_vars, terminator.arguments):
                # `idx` is the original index of the yielded var, subtract 1 as the first block arg is the loop var
                idx = op.body.block.args.index(yielded_var) - 1
                csl.StoreVarOp(iter_var, load_vars[idx])
            csl.ActivateOp(cond_task_id, csl.TaskKind.LOCAL)
            csl.ReturnOp()

        # for-loop body func
        with ImplicitBuilder(body_func.body.block):
            body_vars = [csl.LoadVarOp(var) for var in [iv, *iter_vars]]
            for v in body_vars:
                v.res.name_hint = f"{v.var.name_hint}_bdy"
        rewriter.inline_block(
            op.body.block,
            InsertPoint.at_end(body_func.body.block),
            [v.res for v in body_vars],
        )
        rewriter.insert_op(
            csl.CallOp(SymbolRefAttr(inc_func.sym_name)), InsertPoint.before(terminator)
        )
        rewriter.replace_op(terminator, csl.ReturnOp())

        # place funcs and erase now-empty for-loop
        rewriter.insert_op(
            [cond_func, body_func, inc_func, post_func], InsertPoint.after(parent_func)
        )
        rewriter.erase_matched_op()

    @staticmethod
    def _is_inside_wrapper_outside_apply(op: Operation):
        """Returns if the op is inside `csl_wrapper.module` and contains a `csl_stencil.apply`."""
        is_inside_wrapper = False
        is_inside_apply = False
        has_apply_inside = False

        parent_op = op.parent_op()
        while parent_op:
            if isinstance(parent_op, csl_wrapper.ModuleOp):
                is_inside_wrapper = True
            elif isinstance(parent_op, csl_stencil.ApplyOp):
                is_inside_apply = True
            parent_op = parent_op.parent_op()

        for child_op in op.walk():
            if isinstance(child_op, csl_stencil.ApplyOp):
                has_apply_inside = True
                break

        return is_inside_wrapper and not is_inside_apply and has_apply_inside


@dataclass(frozen=True)
class CopyArithConstants(RewritePattern):
    """ """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: arith.ConstantOp, rewriter: PatternRewriter, /):
        if not (parent_func := self._get_enclosing_function(op)):
            return
        for use in list(op.result.uses):
            use_func = self._get_enclosing_function(use.operation)
            if use_func != parent_func:
                rewriter.insert_op(cln := op.clone(), InsertPoint.before(use.operation))
                op.result.replace_by_if(cln.result, lambda x: x == use)

    @staticmethod
    def _get_enclosing_function(op: Operation) -> csl.FuncOp | None:
        parent = op.parent_op()
        while parent and not isinstance(parent, csl.FuncOp):
            parent = parent.parent_op()
        return parent


@dataclass(frozen=True)
class CslStencilHandleAsyncControlFlow(ModulePass):
    """
    Handles the async control flow of csl_stencil.apply and any enclosing loops
    by translating control flow into a csl.func call graph.
    """

    name = "csl-stencil-handle-async-flow"

    task_ids: tuple[int, ...]
    """
    Available task IDs that this pass is free to allocate.
    """

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        module_pass = PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    ConvertForLoopToCallGraphPass(0, list(self.task_ids)),
                    HandleCslStencilApplyAsyncCF(0),
                ]
            ),
            apply_recursively=False,
        )
        module_pass.rewrite_module(op)
        PatternRewriteWalker(
            CopyArithConstants(), apply_recursively=False
        ).rewrite_module(op)



from dataclasses import dataclass
from typing import cast

from xdsl.context import Context
from xdsl.dialects import riscv, riscv_snitch, snitch, snitch_stream
from xdsl.dialects.builtin import ModuleOp
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    GreedyRewritePatternApplier,
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


class AllocateSnitchStreamingRegionRegisters(RewritePattern):
    """
    Allocates the registers in the body of a `snitch_stream.streaming_region` operation by
    assigning them to the ones specified by the streams.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: snitch_stream.StreamingRegionOp, rewriter: PatternRewriter, /
    ):
        block = op.body.block

        for index, input_stream in enumerate(block.args):
            rewriter.replace_value_with_new_type(
                input_stream, snitch.ReadableStreamType(riscv.Registers.FT[index])
            )

        input_count = len(op.inputs)

        for index, output_stream in enumerate(block.args[input_count:]):
            rewriter.replace_value_with_new_type(
                output_stream,
                snitch.WritableStreamType(riscv.Registers.FT[index + input_count]),
            )


class AllocateRiscvSnitchReadRegisters(RewritePattern):
    """
    Propagates the register allocation done at the stream level to the values read from
    the streams.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: riscv_snitch.ReadOp, rewriter: PatternRewriter, /):
        stream_type = cast(
            snitch.ReadableStreamType[riscv.FloatRegisterType], op.stream.type
        )
        rewriter.replace_value_with_new_type(op.res, stream_type.element_type)


class AllocateRiscvSnitchWriteRegisters(RewritePattern):
    """
    Propagates the register allocation done at the stream level to the values written to
    the streams.
    """

    @op_type_rewrite_pattern
    def match_and_rewrite(self, op: riscv_snitch.WriteOp, rewriter: PatternRewriter, /):
        stream_type = cast(
            snitch.WritableStreamType[riscv.FloatRegisterType], op.stream.type
        )
        rewriter.replace_value_with_new_type(op.value, stream_type.element_type)


@dataclass(frozen=True)
class SnitchRegisterAllocation(ModulePass):
    """
    Allocates unallocated registers for snitch operations.
    """

    name = "snitch-allocate-registers"

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        PatternRewriteWalker(
            AllocateSnitchStreamingRegionRegisters(),
            apply_recursively=False,
        ).rewrite_module(op)
        PatternRewriteWalker(
            GreedyRewritePatternApplier(
                [
                    AllocateRiscvSnitchReadRegisters(),
                    AllocateRiscvSnitchWriteRegisters(),
                ]
            ),
            apply_recursively=False,
        ).rewrite_module(op)



#!/usr/bin/env python3
"""Utilities for profiling ASV benchmarks with a variety of tools."""

import cProfile
import subprocess
import time
from argparse import ArgumentParser, Namespace
from collections.abc import Callable
from pathlib import Path
from typing import Any, NamedTuple, cast

from benchmarks.warmed_timeit import warmed_timeit

DEFAULT_OUTPUT_DIRECTORY = Path(__file__).parent / "profiles"
PROFILERS = (
    "run",
    "timeit",
    "snakeviz",
    "viztracer",
    "flameprof",
    "pyinstrument",
    "dis",
)


class Benchmark(NamedTuple):
    """A wrapper for a benchmark function with optional setup funtion."""

    body: Callable[[], Any]
    setup: Callable[[], Any] | None = None


def parse_arguments(benchmark_names: list[str]) -> ArgumentParser:
    """Parse the arguments for the profiler tool."""
    parser = ArgumentParser()

    parser.add_argument(
        "test",
        choices=benchmark_names + ["all"],
        help="the name of the benchmark to run, `all` to run all benchmarks",
    )
    parser.add_argument(
        "profiler",
        choices=PROFILERS,
        nargs="?",
        default=PROFILERS[0],
        help="the profiler to use",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=Path,
        default=DEFAULT_OUTPUT_DIRECTORY,
        help="the directory into which to write out the profile files",
    )
    parser.add_argument(
        "-q", "--quiet", action="store_true", help="don't show the profiler's UI"
    )

    return parser


def get_benchmark_runs(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
) -> list[tuple[str, Benchmark]]:
    """Get the benchmark to profile."""
    if args.test == "all":
        return list(benchmarks.items())

    name = args.test
    test = benchmarks[name]
    return [(name, test)]


def run_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
    warmup: bool = False,
) -> None:
    """Directly run a benchmark."""
    benchmark_runs = get_benchmark_runs(args, benchmarks)
    for name, (test, setup) in benchmark_runs:
        if warmup:
            if setup is not None:
                setup()
            test()
        if setup is not None:
            setup()
        start_time = time.time()
        test()
        print(f"Test {name} ran in: {time.time() - start_time:.5f}s")


def timeit_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
) -> None:
    """Use a custom function based on timeit to run a benchmark."""
    benchmark_runs = get_benchmark_runs(args, benchmarks)
    for name, (test, setup) in benchmark_runs:
        me, _, std = warmed_timeit(test, setup=setup)
        print(f"Test {name} ran in: {me:.3g} ± {std:.3g}s")


def cprofile_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
    warmup: bool = False,
) -> Path:
    """Use cProfile to profile a benchmark."""
    benchmark_runs = get_benchmark_runs(args, benchmarks)
    if len(benchmark_runs) != 1:
        raise ValueError("Cannot profile multiple benchmarks together")
    name, (test, setup) = benchmark_runs[0]
    output_prof = args.output / f"{name}.prof"
    if warmup:
        if setup is not None:
            setup()
        test()
    if setup is not None:
        setup()
    profiler = cProfile.Profile()
    profiler.enable()
    test()
    profiler.disable()
    profiler.dump_stats(str(output_prof))
    return output_prof


def viztracer_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
    warmup: bool = True,
    duration: float | None = 0,
) -> Path:
    """Use VizTracer to profile a benchmark."""
    from viztracer import VizTracer  # pyright: ignore[reportMissingTypeStubs]

    benchmark_runs = get_benchmark_runs(args, benchmarks)
    if len(benchmark_runs) != 1:
        raise ValueError("Cannot profile multiple benchmarks together")
    name, (test, setup) = benchmark_runs[0]
    output_prof = args.output / f"{name}.json"
    if warmup:
        if setup is not None:
            setup()
        test()
    if setup is not None:
        setup()

    def wrap() -> None:
        test()

    def fix_time(duration: float) -> float:
        wrap()
        while (end := time.perf_counter()) - start < duration:
            pass
        return end

    if duration is not None:
        tracer = VizTracer(output_file=str(output_prof))
        start = time.perf_counter()
        tracer.start()
        _end = fix_time(duration)
        tracer.save()
    else:
        with VizTracer(output_file=str(output_prof)):
            test()
    return output_prof


def pyinstrument_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
    warmup: bool = True,
) -> Path:
    """Use pyinstrument to profile a benchmark."""
    from pyinstrument import Profiler

    benchmark_runs = get_benchmark_runs(args, benchmarks)
    if len(benchmark_runs) != 1:
        raise ValueError("Cannot profile multiple benchmarks together")
    name, (test, setup) = benchmark_runs[0]
    output_prof = args.output / f"{name}.html"
    if warmup:
        if setup is not None:
            setup()
        test()
    if setup is not None:
        setup()
    profiler = Profiler(interval=1e-9)
    profiler.start()
    test()
    profiler.stop()
    profiler.write_html(output_prof)
    return output_prof


def dis_benchmark(
    args: Namespace,
    benchmarks: dict[str, Benchmark],
):
    """Use dis to disassemble a benchmark."""
    from bytesight import profile_bytecode

    benchmark_runs = get_benchmark_runs(args, benchmarks)
    if len(benchmark_runs) != 1:
        raise ValueError("Cannot disassemble multiple benchmarks together")
    _, (test, setup) = benchmark_runs[0]
    if setup is not None:
        setup()
    profile_bytecode(test)


def show(
    args: Namespace,
    output_prof: Path,
    tool: str,
    options: tuple[str, ...] | None = None,
) -> None:
    """Show the profile using the specified tool."""
    if args.quiet:
        return
    if options is None:
        options = cast(tuple[str], ())
    command = ["uv", "run", tool, output_prof, *options]
    subprocess.run(command, check=True)  # noqa: S603


def profile(
    benchmarks: dict[str, Benchmark],
    argv: list[str] | None = None,
) -> None:
    """Run the selected profiler."""
    if not benchmarks:
        raise ValueError("At least one benchmark must be provided to profile!")

    args = parse_arguments(list(benchmarks.keys())).parse_args(args=argv)

    match args.profiler:
        case "run":
            run_benchmark(args, benchmarks)
        case "timeit":
            timeit_benchmark(args, benchmarks)
        case "snakeviz":
            output_prof = cprofile_benchmark(args, benchmarks)
            show(args, output_prof, tool="snakeviz")
        case "flameprof":
            output_prof = cprofile_benchmark(args, benchmarks)
            show(args, output_prof, tool="flameprof")
        case "viztracer":
            output_prof = viztracer_benchmark(args, benchmarks)
            show(args, output_prof, tool="vizviewer")
        case "pyinstrument":
            output_prof = pyinstrument_benchmark(args, benchmarks)
            show(args, output_prof, tool="open")
        case "dis":
            dis_benchmark(args, benchmarks)
        case _:
            raise ValueError("Invalid command!")



"""
Toy language dialect from MLIR tutorial.
"""

from __future__ import annotations

from collections.abc import Sequence
from typing import TypeAlias

from xdsl.dialects.builtin import (
    DenseIntOrFPElementsAttr,
    Float64Type,
    FunctionType,
    StringAttr,
    SymbolNameConstraint,
    SymbolRefAttr,
    TensorType,
    UnrankedTensorType,
    f64,
)
from xdsl.ir import (
    Attribute,
    Block,
    Dialect,
    Operation,
    OpTraits,
    Region,
    SSAValue,
)
from xdsl.irdl import (
    IRDLOperation,
    attr_def,
    irdl_op_definition,
    operand_def,
    opt_attr_def,
    opt_operand_def,
    region_def,
    result_def,
    traits_def,
    var_operand_def,
    var_result_def,
)
from xdsl.pattern_rewriter import (
    PatternRewriter,
    RewritePattern,
)
from xdsl.traits import (
    CallableOpInterface,
    HasCanonicalizationPatternsTrait,
    HasParent,
    HasShapeInferencePatternsTrait,
    IsTerminator,
    Pure,
    SymbolOpInterface,
)
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa

TensorTypeF64: TypeAlias = TensorType[Float64Type]
UnrankedTensorTypeF64: TypeAlias = UnrankedTensorType[Float64Type]
AnyTensorTypeF64: TypeAlias = TensorTypeF64 | UnrankedTensorTypeF64


@irdl_op_definition
class ConstantOp(IRDLOperation):
    """
    Constant operation turns a literal into an SSA value. The data is attached
    to the operation as an attribute. For example:

    ```mlir
      %0 = toy.constant dense<[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]>
                        : tensor<2x3xf64>
    ```
    """

    name = "toy.constant"
    value = attr_def(DenseIntOrFPElementsAttr[Float64Type])
    res = result_def(TensorTypeF64)

    traits = traits_def(Pure())

    def __init__(self, value: DenseIntOrFPElementsAttr):
        super().__init__(result_types=[value.type], attributes={"value": value})

    @staticmethod
    def from_list(data: list[float], shape: list[int]) -> ConstantOp:
        value = DenseIntOrFPElementsAttr.from_list(TensorType(f64, shape), data)
        return ConstantOp(value)

    @staticmethod
    def from_value(value: float) -> ConstantOp:
        return ConstantOp(
            DenseIntOrFPElementsAttr.from_list(TensorType(f64, []), (value,))
        )

    def verify_(self) -> None:
        if not self.res.type == self.value.type:
            raise VerifyException(
                "Expected value and result types to be equal: "
                f"{self.res.type}, {self.value.type}"
            )


class InferAddOpShapeInferencePattern(RewritePattern):
    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter, /):
        if not isinstance(op, AddOp):
            return

        if not (
            isa(op_lhs_type := op.lhs.type, TensorType)
            and isinstance(op_rhs_type := op.rhs.type, TensorType)
        ):
            return
        assert op_lhs_type.get_shape() == op_rhs_type.get_shape()
        if isinstance(op_res_type := op.res.type, TensorType):
            assert op_lhs_type.get_shape() == op_res_type.get_shape()
        else:
            rewriter.replace_value_with_new_type(op.res, op_lhs_type)


class InferAddOpHasShapeInferencePatternsTrait(HasShapeInferencePatternsTrait):
    @classmethod
    def get_shape_inference_patterns(cls) -> tuple[RewritePattern, ...]:
        return (InferAddOpShapeInferencePattern(),)


@irdl_op_definition
class AddOp(IRDLOperation):
    """
    The "add" operation performs element-wise addition between two tensors.
    The shapes of the tensor operands are expected to match.
    """

    name = "toy.add"
    lhs = operand_def(AnyTensorTypeF64)
    rhs = operand_def(AnyTensorTypeF64)
    res = result_def(AnyTensorTypeF64)

    traits = traits_def(Pure(), InferAddOpHasShapeInferencePatternsTrait())

    def __init__(self, lhs: SSAValue, rhs: SSAValue):
        if isa(lhs.type, TensorTypeF64):
            result_type = lhs.type
        else:
            result_type = rhs.type
        super().__init__(result_types=[result_type], operands=[lhs, rhs])

    def verify_(self):
        args = [self.lhs, self.rhs]

        shape = None
        for arg in args:
            # Expect shapes to be the same whenever they are defined, no check for unranked
            if isinstance(arg_type := arg.type, TensorType):
                if shape is None:
                    shape = arg_type.shape
                else:
                    if shape != arg_type.shape:
                        raise VerifyException(
                            "Expected AddOp args to have the same shape"
                        )


class FuncOpCallableInterface(CallableOpInterface):
    @classmethod
    def get_callable_region(cls, op: Operation) -> Region:
        assert isinstance(op, FuncOp)
        return op.body

    @classmethod
    def get_argument_types(cls, op: Operation) -> tuple[Attribute, ...]:
        assert isinstance(op, FuncOp)
        return op.function_type.inputs.data

    @classmethod
    def get_result_types(cls, op: Operation) -> tuple[Attribute, ...]:
        assert isinstance(op, FuncOp)
        return op.function_type.outputs.data


@irdl_op_definition
class FuncOp(IRDLOperation):
    """
    The "toy.func" operation represents a user defined function. These are
    callable SSA-region operations that contain toy computations.

    Example:

    ```mlir
    toy.func @main() {
      %0 = toy.constant dense<5.500000e+00> : tensor<f64>
      %1 = toy.reshape(%0 : tensor<f64>) to tensor<2x2xf64>
      toy.print %1 : tensor<2x2xf64>
      toy.return
    }
    ```
    """

    name = "toy.func"
    body = region_def("single_block")
    sym_name = attr_def(SymbolNameConstraint())
    function_type = attr_def(FunctionType)
    sym_visibility = opt_attr_def(StringAttr)

    traits = traits_def(SymbolOpInterface(), FuncOpCallableInterface())

    def __init__(
        self,
        name: str,
        ftype: FunctionType,
        region: Region | type[Region.DEFAULT] = Region.DEFAULT,
        /,
        private: bool = False,
    ):
        attributes: dict[str, Attribute] = {
            "sym_name": StringAttr(name),
            "function_type": ftype,
        }
        if not isinstance(region, Region):
            region = Region(Block(arg_types=ftype.inputs))
        if private:
            attributes["sym_visibility"] = StringAttr("private")

        return super().__init__(attributes=attributes, regions=[region])

    def verify_(self):
        block = self.body.block

        if not block.ops:
            raise VerifyException("Expected FuncOp to not be empty")

        last_op = block.last_op

        if not isinstance(last_op, ReturnOp):
            raise VerifyException("Expected last op of FuncOp to be a ReturnOp")


@irdl_op_definition
class GenericCallOp(IRDLOperation):
    name = "toy.generic_call"
    arguments = var_operand_def()
    callee = attr_def(SymbolRefAttr)

    res = var_result_def(AnyTensorTypeF64)

    def __init__(
        self,
        callee: str | SymbolRefAttr,
        operands: Sequence[SSAValue],
        return_types: Sequence[Attribute],
    ):
        if isinstance(callee, str):
            callee = SymbolRefAttr(callee)

        return super().__init__(
            operands=[operands],
            result_types=[return_types],
            attributes={"callee": callee},
        )


class MulOpInferShapeInferencePattern(RewritePattern):
    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter, /):
        if not isinstance(op, MulOp):
            return

        if not (
            isa(op_lhs_type := op.lhs.type, TensorType)
            and isinstance(op_rhs_type := op.rhs.type, TensorType)
        ):
            return

        assert op_lhs_type.get_shape() == op_rhs_type.get_shape()
        if isinstance(op_res_type := op.res.type, TensorType):
            assert op_lhs_type.get_shape() == op_res_type.get_shape()
        else:
            rewriter.replace_value_with_new_type(op.res, op_lhs_type)


class InferMulOpHasShapeInferencePatternsTrait(HasShapeInferencePatternsTrait):
    @classmethod
    def get_shape_inference_patterns(cls) -> tuple[RewritePattern, ...]:
        return (MulOpInferShapeInferencePattern(),)


@irdl_op_definition
class MulOp(IRDLOperation):
    """
    The "mul" operation performs element-wise multiplication between two
    tensors. The shapes of the tensor operands are expected to match.
    """

    name = "toy.mul"
    lhs = operand_def(AnyTensorTypeF64)
    rhs = operand_def(AnyTensorTypeF64)
    res = result_def(AnyTensorTypeF64)

    traits = traits_def(Pure(), InferMulOpHasShapeInferencePatternsTrait())

    def __init__(self, lhs: SSAValue, rhs: SSAValue):
        super().__init__(result_types=[lhs.type], operands=[lhs, rhs])

    def verify_(self):
        args = [self.lhs, self.rhs]

        shape = None
        for arg in args:
            # Expect shapes to be the same whenever they are defined, no check for unranked
            if isinstance(arg_type := arg.type, TensorType):
                if shape is None:
                    shape = arg_type.shape
                else:
                    if shape != arg_type.shape:
                        raise VerifyException(
                            "Expected MulOp args to have the same shape"
                        )


@irdl_op_definition
class PrintOp(IRDLOperation):
    """
    The "print" builtin operation prints a given input tensor, and produces
    no results.
    """

    name = "toy.print"
    input = operand_def()

    def __init__(self, input: SSAValue):
        return super().__init__(operands=[input])


@irdl_op_definition
class ReturnOp(IRDLOperation):
    """
    The "return" operation represents a return operation within a function.
    The operation takes an optional tensor operand and produces no results.
    The operand type must match the signature of the function that contains
    the operation. For example:

    ```mlir
      func @foo() -> tensor<2xf64> {
        ...
        toy.return %0 : tensor<2xf64>
      }
    ```
    """

    name = "toy.return"
    input = opt_operand_def(AnyTensorTypeF64)

    traits = traits_def(IsTerminator(), HasParent(FuncOp))

    def __init__(self, input: SSAValue | None = None):
        return super().__init__(operands=[input])

    def verify_(self) -> None:
        func_op = self.parent_op()
        assert isinstance(func_op, FuncOp)

        function_return_types = func_op.function_type.outputs.data
        if function_return_types != tuple(self.operand_types):
            raise VerifyException(
                "Expected arguments to have the same types as the function output types"
            )


class ReshapeOpHasCanonicalizationPatternsTrait(HasCanonicalizationPatternsTrait):
    @classmethod
    def get_canonicalization_patterns(cls) -> tuple[RewritePattern, ...]:
        from ..rewrites.optimise_toy import (
            FoldConstantReshapeOpPattern,
            ReshapeReshapeOpPattern,
        )

        return (ReshapeReshapeOpPattern(), FoldConstantReshapeOpPattern())


@irdl_op_definition
class ReshapeOp(IRDLOperation):
    """
    Reshape operation is transforming its input tensor into a new tensor with
    the same number of elements but different shapes. For example:

    ```mlir
       %0 = toy.reshape (%arg1 : tensor<10xf64>) to tensor<5x2xf64>
    ```
    """

    name = "toy.reshape"
    arg = operand_def(AnyTensorTypeF64)
    res = result_def(TensorTypeF64)

    traits = traits_def(Pure(), ReshapeOpHasCanonicalizationPatternsTrait())

    def __init__(self, arg: SSAValue, result_type: TensorTypeF64 | Sequence[int]):
        if not isinstance(result_type, TensorType):
            result_type = TensorType(f64, result_type)
        return super().__init__(result_types=[result_type], operands=[arg])


class TransposeOpInferShapeInferencePattern(RewritePattern):
    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter, /):
        if not isinstance(op, TransposeOp):
            return

        if not isinstance(op_arg_type := op.arg.type, TensorType):
            return

        arg_shape = op_arg_type.get_shape()
        res_shape = arg_shape[::-1]

        if isinstance(op_res_type := op.res.type, TensorType):
            assert res_shape == op_res_type.get_shape()
        else:
            rewriter.replace_value_with_new_type(op.res, TensorType(f64, res_shape))


class TransposeOpHasShapeInferencePatternsTrait(HasShapeInferencePatternsTrait):
    @classmethod
    def get_shape_inference_patterns(cls) -> tuple[RewritePattern, ...]:
        return (TransposeOpInferShapeInferencePattern(),)


class TransposeOpHasCanonicalizationPatternsTrait(HasCanonicalizationPatternsTrait):
    @classmethod
    def get_canonicalization_patterns(cls) -> tuple[RewritePattern, ...]:
        from ..rewrites.optimise_toy import SimplifyRedundantTranspose

        return (SimplifyRedundantTranspose(),)


@irdl_op_definition
class TransposeOp(IRDLOperation):
    name = "toy.transpose"
    arg = operand_def(AnyTensorTypeF64)
    res = result_def(AnyTensorTypeF64)

    traits = OpTraits(
        lambda: (
            Pure(),
            TransposeOpHasShapeInferencePatternsTrait(),
            TransposeOpHasCanonicalizationPatternsTrait(),
        )
    )

    def __init__(self, arg: SSAValue):
        if isa(arg.type, TensorTypeF64):
            output_type = TensorType(
                arg.type.element_type, reversed(arg.type.get_shape())
            )
        else:
            output_type = arg.type

        super().__init__(operands=[arg], result_types=[output_type])


class CastOpInferShapeInferencePattern(RewritePattern):
    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter, /):
        if not isinstance(op, CastOp):
            return

        if not isinstance(op_arg_type := op.arg.type, TensorType):
            return

        shape = op_arg_type.get_shape()

        if isinstance(op_res_type := op.res.type, TensorType):
            assert shape == op_res_type.get_shape()
            rewriter.replace_matched_op((), (op.arg,))
        else:
            rewriter.replace_value_with_new_type(op.res, TensorType(f64, shape))


class CastOpHasShapeInferencePatternsTrait(HasShapeInferencePatternsTrait):
    @classmethod
    def get_shape_inference_patterns(cls) -> tuple[RewritePattern, ...]:
        return (CastOpInferShapeInferencePattern(),)


@irdl_op_definition
class CastOp(IRDLOperation):
    name = "toy.cast"
    arg = operand_def(AnyTensorTypeF64)
    res = result_def(AnyTensorTypeF64)

    traits = traits_def(Pure(), CastOpHasShapeInferencePatternsTrait())

    def __init__(self, arg: SSAValue, res: AnyTensorTypeF64 | None = None):
        if res is None:
            res = UnrankedTensorType(f64)

        return super().__init__(
            operands=[arg],
            result_types=[res],
        )


Toy = Dialect(
    "toy",
    [
        ConstantOp,
        AddOp,
        FuncOp,
        GenericCallOp,
        PrintOp,
        MulOp,
        ReturnOp,
        ReshapeOp,
        TransposeOp,
        CastOp,
    ],
    [],
)



"""
Test the definition of attributes and their constraints.
"""

from __future__ import annotations

import re
from collections.abc import Sequence
from dataclasses import dataclass
from enum import auto
from io import StringIO
from typing import Annotated, ClassVar, Generic, TypeAlias

import pytest
from typing_extensions import TypeVar, override

from xdsl.context import Context
from xdsl.dialects.builtin import (
    IndexType,
    IntAttr,
    IntegerAttr,
    IntegerType,
    NoneAttr,
    Signedness,
    StringAttr,
    i32,
)
from xdsl.ir import (
    Attribute,
    AttributeCovT,
    AttributeInvT,
    BitEnumAttribute,
    BuiltinAttribute,
    Data,
    EnumAttribute,
    ParametrizedAttribute,
    SpacedOpaqueSyntaxAttribute,
    StrEnum,
    TypedAttribute,
)
from xdsl.irdl import (
    AllOf,
    AnyAttr,
    AnyOf,
    AttrConstraint,
    BaseAttr,
    ConstraintContext,
    GenericData,
    MessageConstraint,
    ParamAttrConstraint,
    ParamAttrDef,
    ParamDef,
    TypeVarConstraint,
    VarConstraint,
    base,
    irdl_attr_definition,
    param_def,
)
from xdsl.parser import AttrParser, Parser
from xdsl.printer import Printer
from xdsl.utils.exceptions import PyRDLAttrDefinitionError, VerifyException
from xdsl.utils.hints import isa


def test_wrong_attribute_type():
    with pytest.raises(
        TypeError,
        match="Class AbstractAttribute should either be a subclass of 'Data' or 'ParametrizedAttribute'",
    ):

        @irdl_attr_definition
        class AbstractAttribute(Attribute):  # pyright: ignore[reportUnusedClass]
            name = "test.wrong"
            pass


################################################################################
# Data attributes
################################################################################


@irdl_attr_definition
class BoolData(Data[bool]):
    """An attribute holding a boolean value."""

    name = "test.bool"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> bool:
        with parser.in_angle_brackets():
            if parser.parse_optional_keyword("True"):
                return True
            if parser.parse_optional_keyword("False"):
                return False
            parser.raise_error("Expected True or False literal")

    def print_parameter(self, printer: Printer):
        with printer.in_angle_brackets():
            printer.print_string(str(self.data))


@irdl_attr_definition
class IntData(Data[int]):
    """An attribute holding an integer value."""

    name = "test.int"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> int:
        with parser.in_angle_brackets():
            return parser.parse_integer()

    def print_parameter(self, printer: Printer):
        with printer.in_angle_brackets():
            printer.print_int(self.data)


@irdl_attr_definition
class StringData(Data[str]):
    """An attribute holding a string value."""

    name = "test.str"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> str:
        with parser.in_angle_brackets():
            return parser.parse_str_literal()

    def print_parameter(self, printer: Printer):
        with printer.in_angle_brackets():
            printer.print_string(self.data)


def test_simple_data():
    """Test that the definition of a data with a class parameter."""
    b = BoolData(True)
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(b)
    assert stream.getvalue() == "#test.bool<True>"


@irdl_attr_definition
class IntListData(Data[tuple[int, ...]]):
    """
    An attribute holding a list of integers.
    """

    name = "test.int_list"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> tuple[int]:
        raise NotImplementedError()

    def print_parameter(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            with printer.in_square_brackets():
                printer.print_list(self.data, printer.print_int)


def test_non_class_data():
    """Test the definition of a Data with a non-class parameter."""
    attr = IntListData((0, 1, 42))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.int_list<[0, 1, 42]>"


class TestEnum(StrEnum):
    Yes = auto()
    No = auto()
    Maybe = auto()


class TestNonIdentifierEnum(StrEnum):
    """
    The value defined by this StrEnum is not parsable as an identifier, because of the
    contained space.
    While valid as a StrEnum, it is thus invalid to use it in an EnumAttribute.
    """

    Spaced = "left right"


@irdl_attr_definition
class EnumData(EnumAttribute[TestEnum], SpacedOpaqueSyntaxAttribute):
    name = "test.enum"


def test_enum_attribute():
    """Test the definition of an EnumAttribute."""
    attr = EnumData(TestEnum.No)
    assert str(attr) == "#test<enum no>"


def test_indirect_enum_guard():
    EnumType = TypeVar("EnumType", bound=StrEnum)
    with pytest.raises(
        TypeError, match="Only direct inheritance from EnumAttribute is allowed."
    ):

        class IndirectEnumData(  # pyright: ignore[reportUnusedClass]
            EnumAttribute[EnumType]
        ):
            name = "test.indirect_enum"


def test_attribute_def_with_non_identifier_enum():
    """
    Test the definition of an EnumAttribute with a non-identifier enum is
    allowed.
    """

    @irdl_attr_definition
    class TestNonIdentifierEnumAttr(  # pyright: ignore[reportUnusedClass]
        EnumAttribute[TestNonIdentifierEnum]
    ):
        name = "test.non_identifier_enum"


@irdl_attr_definition(init=False)
class BitEnumData(BitEnumAttribute[TestEnum]):
    name = "test.bitenum"
    all_value = "all"
    none_value = "none"


@pytest.mark.parametrize(
    "input,output",
    [
        (None, "#test.bitenum<none>"),
        ([], "#test.bitenum<none>"),
        ([TestEnum.No], "#test.bitenum<no>"),
        ([TestEnum.Yes], "#test.bitenum<yes>"),
        ([TestEnum.No, TestEnum.Yes], "#test.bitenum<yes,no>"),
        ([TestEnum.No, TestEnum.Yes, TestEnum.Maybe], "#test.bitenum<all>"),
        ("all", "#test.bitenum<all>"),
        ("none", "#test.bitenum<none>"),
    ],
)
def test_bit_enum_attribute(input: Sequence[TestEnum] | str, output: str):
    attr = BitEnumData(input)
    assert str(attr) == output


def test_bit_enum_invalid_str():
    with pytest.raises(
        TypeError,
        match="expected string parameter to be one of none or all, got helloworld",
    ):
        BitEnumData("helloworld")


################################################################################
# Typed Attribute
################################################################################


def test_typed_attribute():
    with pytest.raises(
        PyRDLAttrDefinitionError,
        match="TypedAttribute TypedAttr should have a 'type' parameter.",
    ):

        @irdl_attr_definition
        class TypedAttr(  # pyright: ignore[reportUnusedClass]
            TypedAttribute
        ):
            name = "test.typed"


def test_typed_attribute_parsing_printing():
    """
    Test that non builtin TypedAttributes are parsed and printed correctly.
    """

    @irdl_attr_definition
    class TypedAttr(TypedAttribute):
        name = "test.typed"
        value: IntAttr
        type: IntegerType

        @classmethod
        def parse_parameters(cls, parser: AttrParser) -> Sequence[Attribute]:
            with parser.in_angle_brackets():
                value = parser.parse_integer()
            parser.parse_punctuation(":")
            type = parser.parse_type()
            return (IntAttr(value), type)

        def print_parameters(self, printer: Printer) -> None:
            with printer.in_angle_brackets():
                printer.print_int(self.value.data)
            printer.print_string(" : ")
            printer.print_attribute(self.type)

        @classmethod
        def get_type_index(cls) -> int:
            return 1

        @staticmethod
        def parse_with_type(
            parser: AttrParser,
            type: Attribute,
        ) -> TypedAttribute:
            raise NotImplementedError()

        def print_without_type(self, printer: Printer) -> None:
            raise NotImplementedError()

    ctx = Context()
    ctx.load_attr_or_type(TypedAttr)
    attr = Parser(ctx, "#test.typed<42> : i32").parse_attribute()
    assert attr == TypedAttr(IntAttr(42), i32)

    assert str(attr) == "#test.typed<42> : i32"


################################################################################
# IntegerAttr
################################################################################


def test_signed_integer_attr():
    """Test the verification of a signed integer attribute."""
    with pytest.raises(VerifyException):
        IntegerAttr(1 << 31, IntegerType(32, Signedness.SIGNED))

    with pytest.raises(VerifyException):
        IntegerAttr(-(1 << 31) - 1, IntegerType(32, Signedness.SIGNED))

    IntegerAttr((1 << 31) - 1, IntegerType(32, Signedness.SIGNED))
    IntegerAttr(-(1 << 31), IntegerType(32, Signedness.SIGNED))


def test_unsigned_integer_attr():
    """Test the verification of a unsigned integer attribute."""
    with pytest.raises(VerifyException):
        IntegerAttr(1 << 32, IntegerType(32, Signedness.UNSIGNED))

    with pytest.raises(VerifyException):
        IntegerAttr(-1, IntegerType(32, Signedness.UNSIGNED))

    IntegerAttr((1 << 32) - 1, IntegerType(32, Signedness.UNSIGNED))


def test_signless_integer_attr():
    """Test the verification of a signless integer attribute."""
    with pytest.raises(VerifyException):
        IntegerAttr((1 << 32), IntegerType(32, Signedness.SIGNLESS))

    with pytest.raises(VerifyException):
        IntegerAttr(-(1 << 32) - 1, IntegerType(32, Signedness.SIGNLESS))

    IntegerAttr(1 << 32 - 1, IntegerType(32, Signedness.SIGNLESS))
    IntegerAttr(-(1 << 31), IntegerType(32, Signedness.SIGNLESS))


################################################################################
# PyRDL Base constraints
################################################################################


@irdl_attr_definition
class BoolWrapperAttr(ParametrizedAttribute):
    name = "test.bool_wrapper"

    param: BoolData


def test_bose_constraint():
    """Test the verifier of a base attribute type constraint."""
    attr = BoolWrapperAttr(BoolData(True))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.bool_wrapper<#test.bool<True>>"


def test_base_constraint_fail():
    """Test the verifier of a union constraint."""
    with pytest.raises(
        VerifyException, match="#test.str<foo> should be of base attribute test.bool"
    ):
        BoolWrapperAttr(StringData("foo"))  # pyright: ignore[reportArgumentType]


################################################################################
# PyRDL union constraints
################################################################################


@irdl_attr_definition
class BoolOrIntParamAttr(ParametrizedAttribute):
    name = "test.bool_or_int"

    param: BoolData | IntData


def test_union_constraint_left():
    """Test the verifier of a union constraint."""
    attr = BoolOrIntParamAttr(BoolData(True))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.bool_or_int<#test.bool<True>>"


def test_union_constraint_right():
    """Test the verifier of a union constraint."""
    attr = BoolOrIntParamAttr(IntData(42))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.bool_or_int<#test.int<42>>"


def test_union_constraint_fail():
    """Test the verifier of a union constraint."""
    with pytest.raises(VerifyException, match="Unexpected attribute #test.str<foo>"):
        BoolOrIntParamAttr(StringData("foo"))  # pyright: ignore[reportArgumentType]


################################################################################
# PyRDL Annotated constraints
################################################################################


class PositiveIntConstr(AttrConstraint):
    def verify(self, attr: Attribute, constraint_context: ConstraintContext) -> None:
        if not isinstance(attr, IntData):
            raise VerifyException(
                f"Expected {IntData.name} attribute, but got {attr.name}."
            )
        if attr.data <= 0:
            raise VerifyException(f"Expected positive integer, got {attr.data}.")

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> PositiveIntConstr:
        return self


@irdl_attr_definition
class PositiveIntAttr(ParametrizedAttribute):
    name = "test.positive_int"

    param: Attribute = param_def(PositiveIntConstr())


def test_annotated_constraint():
    """Test the verifier of an annotated constraint."""
    attr = PositiveIntAttr(IntData(42))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.positive_int<#test.int<42>>"


def test_annotated_constraint_fail():
    """Test that the verifier of an annotated constraint can fail."""
    with pytest.raises(VerifyException, match="Expected positive integer, got -42."):
        PositiveIntAttr(IntData(-42))


################################################################################
# PyRDL Generic constraints
################################################################################

_T = TypeVar("_T", bound=BoolData | IntData)


@irdl_attr_definition
class ParamWrapperAttr(Generic[_T], ParametrizedAttribute):
    name = "test.int_or_bool_generic"

    param: _T


def test_typevar_attribute_int():
    """Test the verifier of a generic attribute."""
    attr = ParamWrapperAttr[IntData](IntData(42))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.int_or_bool_generic<#test.int<42>>"


def test_typevar_attribute_bool():
    """Test the verifier of a generic attribute."""
    attr = ParamWrapperAttr[BoolData](BoolData(True))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert stream.getvalue() == "#test.int_or_bool_generic<#test.bool<True>>"


def test_typevar_attribute_fail():
    """Test that the verifier of an generic attribute can fail."""
    with pytest.raises(VerifyException, match="Unexpected attribute #test.str<foo>"):
        ParamWrapperAttr(StringData("foo"))  # pyright: ignore


@irdl_attr_definition
class ParamConstrAttr(ParametrizedAttribute):
    name = "test.param_constr"

    param: ParamWrapperAttr[IntData]


def test_param_attr_constraint():
    """Test the verifier of an attribute with a parametric constraint."""
    attr = ParamConstrAttr(ParamWrapperAttr(IntData(42)))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.param_constr<#test.int_or_bool_generic<#test.int<42>>>"
    )


def test_param_attr_constraint_fail():
    """
    Test that the verifier of an attribute with
    a parametric constraint can fail.
    """
    with pytest.raises(
        VerifyException, match="#test.bool<True> should be of base attribute test.int"
    ):
        ParamConstrAttr(ParamWrapperAttr(BoolData(True)))  # pyright: ignore


_U = TypeVar("_U", bound=IntData)


@irdl_attr_definition
class NestedParamWrapperAttr(Generic[_U], ParametrizedAttribute):
    name = "test.nested_param_wrapper"

    param: ParamWrapperAttr[_U]


def test_nested_generic_constraint():
    """
    Test the verifier of an attribute with a generic
    constraint used in a parametric constraint.
    """
    attr = NestedParamWrapperAttr[IntData](ParamWrapperAttr(IntData(42)))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.nested_param_wrapper<#test.int_or_bool_generic<#test.int<42>>>"
    )


def test_nested_generic_constraint_fail():
    """
    Test that the verifier of an attribute with
    a parametric constraint can fail.
    """
    with pytest.raises(
        VerifyException, match="#test.bool<True> should be of base attribute test.int"
    ):
        NestedParamWrapperAttr(ParamWrapperAttr(BoolData(True)))  # pyright: ignore


@irdl_attr_definition
class NestedParamConstrAttr(ParametrizedAttribute):
    name = "test.nested_param_constr"

    param: NestedParamWrapperAttr[Annotated[IntData, PositiveIntConstr()]]


def test_nested_param_attr_constraint():
    """
    Test the verifier of a nested parametric constraint.
    """
    attr = NestedParamConstrAttr(NestedParamWrapperAttr(ParamWrapperAttr(IntData(42))))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.nested_param_constr<#test.nested_param_wrapper<#test.int_or_bool_generic<#test.int<42>>>>"
    )


def test_nested_param_attr_constraint_fail():
    """
    Test that the verifier of a nested parametric constraint can fail.
    """
    with pytest.raises(VerifyException, match="Expected positive integer, got -42."):
        NestedParamConstrAttr(NestedParamWrapperAttr(ParamWrapperAttr(IntData(-42))))


################################################################################
# Message Constraint
################################################################################


@irdl_attr_definition
class InformativeAttr(ParametrizedAttribute):
    name = "test.informative"

    param: Attribute = param_def(
        MessageConstraint(
            NoneAttr,
            "Dear user, here's what this constraint means in your abstraction.",
        )
    )


def test_informative_attribute():
    okay = InformativeAttr(NoneAttr())
    okay.verify()

    with pytest.raises(
        VerifyException,
        match="Dear user, here's what this constraint means in your abstraction.\nUnderlying verification failure: #test.int<42> should be of base attribute none",
    ):
        InformativeAttr(IntData(42))


def test_informative_constraint():
    """
    Test the verifier of an informative constraint.
    """
    constr = MessageConstraint(NoneAttr(), "User-enlightening message.")
    with pytest.raises(
        VerifyException,
        match="User-enlightening message.\nUnderlying verification failure: Expected attribute none but got #builtin.int<1>",
    ):
        constr.verify(IntAttr(1), ConstraintContext())
    assert constr.can_infer(set())
    assert constr.get_bases() == {NoneAttr}


################################################################################
# GenericData definition
################################################################################


@irdl_attr_definition
class ListData(Generic[AttributeInvT], GenericData[tuple[AttributeInvT, ...]]):
    name = "test.list"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> tuple[AttributeInvT, ...]:
        raise NotImplementedError()

    def print_parameter(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            printer.print_string("[")
            printer.print_list(self.data, printer.print_attribute)
            printer.print_string("]")

    @staticmethod
    @override
    def constr(
        constr: AttrConstraint[AttributeCovT],
    ) -> DataListAttr[AttributeCovT]:
        return DataListAttr(constr)

    @staticmethod
    def from_list(data: list[AttributeInvT]) -> ListData[AttributeInvT]:
        return ListData(tuple(data))


AnyListData: TypeAlias = ListData[Attribute]


@dataclass(frozen=True)
class DataListAttr(AttrConstraint[ListData[AttributeInvT]]):
    """
    A constraint that enforces that the elements of a ListData all respect
    a constraint.
    """

    elem_constr: AttrConstraint[AttributeInvT]

    def verify(
        self,
        attr: Attribute,
        constraint_context: ConstraintContext,
    ) -> None:
        if not isa(attr, ListData):
            raise VerifyException(
                f"Expected {attr} to be instance of {ListData.__name__}"
            )
        for e in attr.data:
            self.elem_constr.verify(e, constraint_context)

    def mapping_type_vars(
        self, type_var_mapping: dict[TypeVar, AttrConstraint]
    ) -> DataListAttr[AttributeInvT]:
        return DataListAttr(self.elem_constr.mapping_type_vars(type_var_mapping))


class Test_generic_data_verifier:
    def test_generic_data_verifier(self):
        """
        Test that a GenericData can be created.
        """
        attr = ListData((BoolData(True), ListData((BoolData(False),))))
        stream = StringIO()
        p = Printer(stream=stream)
        p.print_attribute(attr)
        assert (
            stream.getvalue()
            == "#test.list<[#test.bool<True>, #test.list<[#test.bool<False>]>]>"
        )


@irdl_attr_definition
class ListDataWrapper(ParametrizedAttribute):
    name = "test.list_wrapper"

    val: ListData[BoolData]


def test_generic_data_wrapper_verifier():
    """
    Test that a GenericData used in constraints pass the verifier when correct.
    """
    attr = ListDataWrapper(ListData((BoolData(True), BoolData(False))))
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.list_wrapper<#test.list<[#test.bool<True>, #test.bool<False>]>>"
    )


def test_generic_data_wrapper_verifier_failure():
    """
    Test that a GenericData used in constraints fails
    the verifier when constraints are not satisfied.
    """
    with pytest.raises(
        VerifyException,
        match=re.escape(
            "#test.list<[#test.bool<False>]> should be of base attribute test.bool"
        ),
    ):
        ListDataWrapper(ListData((BoolData(True), ListData((BoolData(False),)))))  # pyright: ignore[reportArgumentType]


@irdl_attr_definition
class ListDataNoGenericsWrapper(ParametrizedAttribute):
    name = "test.list_no_generics_wrapper"

    val: AnyListData


def test_generic_data_no_generics_wrapper_verifier():
    """
    Test that GenericType can be used in constraints without a parameter.
    """
    attr = ListDataNoGenericsWrapper(
        ListData((BoolData(True), ListData((BoolData(False),))))
    )
    stream = StringIO()
    p = Printer(stream=stream)
    p.print_attribute(attr)
    assert (
        stream.getvalue()
        == "#test.list_no_generics_wrapper<#test.list<[#test.bool<True>, #test.list<[#test.bool<False>]>]>>"
    )


################################################################################
# Parametrized attribute definition
################################################################################


@irdl_attr_definition
class ParamAttrDefAttr(ParametrizedAttribute):
    name = "test.param_attr_def_attr"

    arg1: Attribute
    arg2: BoolData

    # Check that we can define methods in attribute definition
    def test(self):
        pass


def test_irdl_definition():
    """Test that we can get the IRDL definition of a parametrized attribute."""
    assert ParamAttrDefAttr.get_irdl_definition() == ParamAttrDef(
        "test.param_attr_def_attr",
        [("arg1", ParamDef(AnyAttr())), ("arg2", ParamDef(BaseAttr(BoolData)))],
    )


@irdl_attr_definition
class ParamAttrDefAttr2(ParametrizedAttribute):
    name = "test.param_attr_def_attr"

    arg1: Attribute = param_def(base(IntAttr))
    arg2: BoolData

    # Check that we can define methods in attribute definition
    def test(self):
        pass


def test_irdl_definition2():
    """Test that we can get the IRDL definition of a parametrized attribute."""

    assert ParamAttrDefAttr2.get_irdl_definition() == ParamAttrDef(
        "test.param_attr_def_attr",
        [
            ("arg1", ParamDef(AnyAttr() & BaseAttr(IntAttr))),
            ("arg2", ParamDef(BaseAttr(BoolData))),
        ],
    )


class InvalidTypedFieldTestAttr(ParametrizedAttribute):
    name = "test.invalid_typed_field"

    x: int


def test_invalid_typed_field():
    """Check that typed fields are not allowed."""
    with pytest.raises(PyRDLAttrDefinitionError):
        irdl_attr_definition(InvalidTypedFieldTestAttr)


class InvalidUntypedFieldTestAttr(ParametrizedAttribute):
    name = "test.invalid_untyped_field"

    x = 2


def test_invalid_field():
    """Check that untyped fields are not allowed."""
    with pytest.raises(PyRDLAttrDefinitionError):
        irdl_attr_definition(InvalidUntypedFieldTestAttr)


@irdl_attr_definition
class OveriddenInitAttr(ParametrizedAttribute):
    name = "test.overidden_init"

    param: Attribute

    def __init__(self, param: int | str):
        match param:
            case int():
                super().__init__(IntData(param))
            case str():
                super().__init__(StringData(param))


def test_generic_constructor():
    """Test the generic constructor of a ParametrizedAttribute."""

    param = IntData(42)
    attr = OveriddenInitAttr.new([param])

    assert isinstance(attr, OveriddenInitAttr)
    assert attr.param == param


def test_custom_constructor():
    """Test the use of custom constructors in ParametrizedAttribute."""

    assert OveriddenInitAttr.new([IntData(42)]) == OveriddenInitAttr(42)
    assert OveriddenInitAttr.new([StringData("17")]) == OveriddenInitAttr("17")


@irdl_attr_definition
class GenericAttr(Generic[AttributeInvT], ParametrizedAttribute):
    name = "test.generic_attr"

    param: AttributeInvT


def test_generic_attr():
    """Test the generic parameter of a ParametrizedAttribute."""

    assert GenericAttr.get_irdl_definition() == ParamAttrDef(
        "test.generic_attr",
        [
            (
                "param",
                ParamDef(
                    TypeVarConstraint(
                        type_var=AttributeInvT,
                        base_constraint=AnyAttr(),
                    )
                ),
            )
        ],
    )

    assert base(GenericAttr[IntAttr]) == ParamAttrConstraint(
        GenericAttr, (BaseAttr(IntAttr),)
    )


################################################################################
# ConstraintVar
################################################################################


@irdl_attr_definition
class ConstraintVarAttr(ParametrizedAttribute):
    name = "test.constraint_var"

    T: ClassVar = VarConstraint("T", BaseAttr(IntegerType))

    param1: IntegerAttr = param_def(IntegerAttr.constr(type=T))
    param2: IntegerAttr = param_def(IntegerAttr.constr(type=T))


def test_constraint_var():
    """Test that ConstraintVar can be used in attributes."""
    ConstraintVarAttr.new([IntegerAttr(42, 32), IntegerAttr(17, 32)])
    ConstraintVarAttr.new([IntegerAttr(42, 64), IntegerAttr(17, 64)])


def test_constraint_var_fail_non_equal():
    """Test that constraint variables must be equal."""
    with pytest.raises(VerifyException):
        ConstraintVarAttr.new([IntegerAttr(42, 32), IntegerAttr(42, 64)])
    with pytest.raises(VerifyException):
        ConstraintVarAttr.new([IntegerAttr(42, 64), IntegerAttr(42, 32)])


def test_constraint_var_fail_not_satisfy_constraint():
    """Test that constraint variables must satisfy the underlying constraint."""
    with pytest.raises(VerifyException):
        ConstraintVarAttr.new(
            [IntegerAttr(42, IndexType()), IntegerAttr(17, IndexType())]
        )


################################################################################
# Names
################################################################################


def test_non_builtin_name_fail():
    """
    Test that the name of an attribute is properly checked
    when it is not a builtin attribute.
    """
    with pytest.raises(PyRDLAttrDefinitionError, match="is not a valid attribute name"):

        @irdl_attr_definition
        class NonBuiltinNameAttr(  # pyright: ignore[reportUnusedClass]
            ParametrizedAttribute
        ):
            name = "vector"


def test_non_builtin_name():
    """
    Test that the name of an attribute is properly checked
    when it is not a builtin attribute.
    """

    @irdl_attr_definition
    class NonBuiltinNameAttr(  # pyright: ignore[reportUnusedClass]
        ParametrizedAttribute
    ):
        name = "test.vector"


def test_builtin_name():
    """
    Test that builtin attribute names are not checked.
    """

    @irdl_attr_definition
    class BuiltinNameAttr(  # pyright: ignore[reportUnusedClass]
        ParametrizedAttribute, BuiltinAttribute
    ):
        name = "builtin.vector"


################################################################################
# Mapping Type Var
################################################################################


@irdl_attr_definition
class A(Data[int]):
    name = "test.a"


@irdl_attr_definition
class B(Data[int]):
    name = "test.b"


@irdl_attr_definition
class C(Data[int]):
    name = "test.c"


_A = TypeVar("_A", bound=Attribute)


def test_var_constraint():
    var_constraint = VarConstraint("var", TypeVarConstraint(_A, BaseAttr(A)))

    with pytest.raises(KeyError, match="Mapping value missing for type var ~_A"):
        var_constraint.mapping_type_vars({})

    assert var_constraint.mapping_type_vars({_A: BaseAttr(B)}) == VarConstraint(
        "var", BaseAttr(B)
    )


def test_typevar_constraint():
    typevar_constraint = TypeVarConstraint(_A, BaseAttr(A))

    with pytest.raises(
        KeyError, match=re.escape("Mapping value missing for type var ~_A")
    ):
        assert typevar_constraint.mapping_type_vars({})
    assert typevar_constraint.mapping_type_vars({_A: BaseAttr(B)}) == BaseAttr(B)


def test_message_constraint():
    message_constraint = MessageConstraint(
        TypeVarConstraint(_A, BaseAttr(A)), "test message"
    )

    assert message_constraint.mapping_type_vars({_A: BaseAttr(B)}) == MessageConstraint(
        BaseAttr(B), "test message"
    )


def test_anyof_constraint():
    anyof_constraint = AnyOf((TypeVarConstraint(_A, BaseAttr(A)), BaseAttr(B)))

    assert anyof_constraint.mapping_type_vars({_A: BaseAttr(C)}) == AnyOf(
        (BaseAttr(C), BaseAttr(B))
    )


def test_allof_constraint():
    allof_constraint = AllOf((TypeVarConstraint(_A, BaseAttr(A)), BaseAttr(B)))

    assert allof_constraint.mapping_type_vars({_A: BaseAttr(C)}) == AllOf(
        (BaseAttr(C), BaseAttr(B))
    )


################################################################################
# Constant ClassVar
################################################################################


def test_class_var_pass():
    """Test that ClassVar constants are allowed in attribute definitions."""

    @irdl_attr_definition
    class ClassVarAttr(ParametrizedAttribute):  # pyright: ignore[reportUnusedClass]
        name = "test.class_var"
        CONSTANT: ClassVar[int]
        param: IntData

    @irdl_attr_definition
    class ClassVarAttr2(ParametrizedAttribute):  # pyright: ignore[reportUnusedClass]
        name = "test.class_var"
        CONSTANT: ClassVar[int] = 2
        param: IntData


def test_class_var_fail():
    """Test that lowercase ClassVar fields are not allowed."""
    with pytest.raises(
        PyRDLAttrDefinitionError,
        match='Invalid ClassVar name "constant", must be uppercase.',
    ):

        @irdl_attr_definition
        class InvalidClassVarAttr(ParametrizedAttribute):  # pyright: ignore[reportUnusedClass]
            name = "test.invalid_class_var"
            constant: ClassVar[int]  # Should be uppercase
            param: IntData


################################################################################
# Converters
################################################################################


@irdl_attr_definition
class ConverterAttr(ParametrizedAttribute):
    name = "test.converters"

    string: StringAttr = param_def(converter=StringAttr.get)

    i: IntAttr = param_def(converter=IntAttr.get)


def test_converters():
    string = "My string"
    string_attr = StringAttr(string)

    i = 2
    i_attr = IntAttr(i)

    attr_no_convertion = ConverterAttr(string_attr, i_attr)
    assert attr_no_convertion.i == i_attr
    assert attr_no_convertion.string == string_attr

    attr_some_convertion = ConverterAttr(string_attr, i)
    assert attr_some_convertion.i == i_attr
    assert attr_some_convertion.string == string_attr
    assert attr_no_convertion == attr_some_convertion

    attr_all_convertion = ConverterAttr(string, i)
    assert attr_all_convertion.i == i_attr
    assert attr_all_convertion.string == string_attr
    assert attr_no_convertion == attr_all_convertion



"""
ARM dialect, based on the ISA [specification](https://developer.arm.com/documentation/102374/0101/Overview).
"""

from typing import IO

from xdsl.backend.assembly_printer import AssemblyPrinter
from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import Dialect

from .ops import CmpRegOp, DSMovOp, DSSMulOp, GetRegisterOp, LabelOp
from .register import IntRegisterType


def print_assembly(module: ModuleOp, output: IO[str]) -> None:
    printer = AssemblyPrinter(stream=output)
    printer.print_module(module)


ARM = Dialect(
    "arm",
    [
        GetRegisterOp,
        CmpRegOp,
        DSMovOp,
        DSSMulOp,
        LabelOp,
    ],
    [
        IntRegisterType,
    ],
)



import pytest

from xdsl.utils.exceptions import ParseError
from xdsl.utils.lexer import Input, Position
from xdsl.utils.mlir_lexer import MLIRLexer, MLIRToken, MLIRTokenKind


def get_token(input: str) -> MLIRToken:
    file = Input(input, "<unknown>")
    lexer = MLIRLexer(file)
    token = lexer.lex()
    return token


def assert_single_token(
    input: str, expected_kind: MLIRTokenKind, expected_text: str | None = None
):
    if expected_text is None:
        expected_text = input

    token = get_token(input)

    assert token.kind == expected_kind
    assert token.text == expected_text


def assert_token_fail(input: str):
    file = Input(input, "<unknown>")
    lexer = MLIRLexer(file)
    with pytest.raises(ParseError):
        lexer.lex()


@pytest.mark.parametrize(
    "text,kind",
    [
        ("->", MLIRTokenKind.ARROW),
        (":", MLIRTokenKind.COLON),
        (",", MLIRTokenKind.COMMA),
        ("...", MLIRTokenKind.ELLIPSIS),
        ("=", MLIRTokenKind.EQUAL),
        (">", MLIRTokenKind.GREATER),
        ("{", MLIRTokenKind.L_BRACE),
        ("(", MLIRTokenKind.L_PAREN),
        ("[", MLIRTokenKind.L_SQUARE),
        ("<", MLIRTokenKind.LESS),
        ("-", MLIRTokenKind.MINUS),
        ("+", MLIRTokenKind.PLUS),
        ("?", MLIRTokenKind.QUESTION),
        ("}", MLIRTokenKind.R_BRACE),
        (")", MLIRTokenKind.R_PAREN),
        ("]", MLIRTokenKind.R_SQUARE),
        ("*", MLIRTokenKind.STAR),
        ("|", MLIRTokenKind.VERTICAL_BAR),
        ("{-#", MLIRTokenKind.FILE_METADATA_BEGIN),
        ("#-}", MLIRTokenKind.FILE_METADATA_END),
    ],
)
def test_punctuation(text: str, kind: MLIRTokenKind):
    assert_single_token(text, kind)


@pytest.mark.parametrize("text", [".", "&", "/"])
def test_punctuation_fail(text: str):
    assert_token_fail(text)


@pytest.mark.parametrize(
    "text", ['""', '"@"', '"foo"', '"\\""', '"\\n"', '"\\\\"', '"\\t"']
)
def test_str_literal(text: str):
    assert_single_token(text, MLIRTokenKind.STRING_LIT)


@pytest.mark.parametrize("text", ['"', '"\\"', '"\\a"', '"\n"', '"\v"', '"\f"'])
def test_str_literal_fail(text: str):
    assert_token_fail(text)


@pytest.mark.parametrize(
    "text", ["a", "A", "_", "a_", "a1", "a1_", "a1_2", "a1_2_3", "a$_.", "a$_.1"]
)
def test_bare_ident(text: str):
    """bare-id ::= (letter|[_]) (letter|digit|[_$.])*"""
    assert_single_token(text, MLIRTokenKind.BARE_IDENT)


@pytest.mark.parametrize(
    "text",
    [
        "@a",
        "@A",
        "@_",
        "@a_",
        "@a1",
        "@a1_",
        "@a1_2",
        "@a1_2_3",
        "@a$_.",
        "@a$_.1",
        '@""',
        '@"@"',
        '@"foo"',
        '@"\\""',
        '@"\\n"',
        '@"\\\\"',
        '@"\\t"',
    ],
)
def test_at_ident(text: str):
    """at-ident ::= `@` (bare-id | string-literal)"""
    assert_single_token(text, MLIRTokenKind.AT_IDENT)


@pytest.mark.parametrize(
    "text",
    ["@", '@"', '@"\\"', '@"\\a"', '@"\n"', '@"\v"', '@"\f"', '@ "a"', "@ f", "@$"],
)
def test_at_ident_fail(text: str):
    """at-ident ::= `@` (bare-id | string-literal)"""
    assert_token_fail(text)


@pytest.mark.parametrize(
    "text", ["0", "1234", "a", "S", "$", "_", ".", "-", "e_.$-324", "e5$-e_", "foo"]
)
def test_prefixed_ident(text: str):
    """hash-ident  ::= `#` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)"""
    """percent-ident  ::= `%` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)"""
    """caret-ident  ::= `^` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)"""
    """exclamation-ident  ::= `!` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)"""
    assert_single_token("#" + text, MLIRTokenKind.HASH_IDENT)
    assert_single_token("%" + text, MLIRTokenKind.PERCENT_IDENT)
    assert_single_token("^" + text, MLIRTokenKind.CARET_IDENT)
    assert_single_token("!" + text, MLIRTokenKind.EXCLAMATION_IDENT)


@pytest.mark.parametrize("text", ["+", '""', "#", "%", "^", "!", "\n", ""])
def test_prefixed_ident_fail(text: str):
    """
    hash-ident  ::= `#` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)
    percent-ident  ::= `%` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)
    caret-ident  ::= `^` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)
    exclamation-ident  ::= `!` (digit+ | (letter|[$._-]) (letter|[$._-]|digit)*)
    """
    assert_token_fail("#" + text)
    assert_token_fail("%" + text)
    assert_token_fail("^" + text)
    assert_token_fail("!" + text)


@pytest.mark.parametrize(
    "text,expected",
    [("0x0", "0"), ("0e", "0"), ("0$", "0"), ("0_", "0"), ("0-", "0"), ("0.", "0")],
)
def test_prefixed_ident_split(text: str, expected: str):
    """Check that the prefixed identifier is split at the right character."""
    assert_single_token("#" + text, MLIRTokenKind.HASH_IDENT, "#" + expected)
    assert_single_token("%" + text, MLIRTokenKind.PERCENT_IDENT, "%" + expected)
    assert_single_token("^" + text, MLIRTokenKind.CARET_IDENT, "^" + expected)
    assert_single_token("!" + text, MLIRTokenKind.EXCLAMATION_IDENT, "!" + expected)


@pytest.mark.parametrize("text", ["0", "01", "123456789", "99", "0x1234", "0xabcdef"])
def test_integer_literal(text: str):
    assert_single_token(text, MLIRTokenKind.INTEGER_LIT)


@pytest.mark.parametrize(
    "text,expected", [("0a", "0"), ("0xg", "0"), ("0xfg", "0xf"), ("0xf.", "0xf")]
)
def test_integer_literal_split(text: str, expected: str):
    assert_single_token(text, MLIRTokenKind.INTEGER_LIT, expected)


@pytest.mark.parametrize(
    "text", ["0.", "1.", "0.2", "38.1243", "92.54e43", "92.5E43", "43.3e-54", "32.E+25"]
)
def test_float_literal(text: str):
    assert_single_token(text, MLIRTokenKind.FLOAT_LIT)


@pytest.mark.parametrize(
    "text,expected", [("3.9e", "3.9"), ("4.5e+", "4.5"), ("5.8e-", "5.8")]
)
def test_float_literal_split(text: str, expected: str):
    assert_single_token(text, MLIRTokenKind.FLOAT_LIT, expected)


@pytest.mark.parametrize("text", ["0", " 0", "   0", "\n0", "\t0", "// Comment\n0"])
def test_whitespace_skip(text: str):
    assert_single_token(text, MLIRTokenKind.INTEGER_LIT, "0")


@pytest.mark.parametrize("text", ["", "   ", "\n\n", "// Comment\n"])
def test_eof(text: str):
    assert_single_token(text, MLIRTokenKind.EOF, "")


@pytest.mark.parametrize(
    "text, expected",
    [
        ("0", 0),
        ("010", 10),
        ("123456789", 123456789),
        ("0x1234", 4660),
        ("0xabcdef23", 2882400035),
    ],
)
def test_token_get_int_value(text: str, expected: int):
    token = get_token(text)
    assert token.kind == MLIRTokenKind.INTEGER_LIT
    assert token.kind.get_int_value(token.span) == expected


@pytest.mark.parametrize(
    "text, expected",
    [
        ("0.", 0.0),
        ("1.", 1.0),
        ("0.2", 0.2),
        ("38.1243", 38.1243),
        ("92.54e43", 92.54e43),
        ("92.5E43", 92.5e43),
        ("43.3e-54", 43.3e-54),
        ("32.E+25", 32.0e25),
    ],
)
def test_token_get_float_value(text: str, expected: float):
    token = get_token(text)
    assert token.kind == MLIRTokenKind.FLOAT_LIT
    assert token.kind.get_float_value(token.span) == expected


@pytest.mark.parametrize(
    "text,expected",
    [
        ('""', ""),
        ('"@"', "@"),
        ('"foo"', "foo"),
        ('"\\""', '"'),
        ('"\\n"', "\n"),
        ('"\\\\"', "\\"),
        ('"\\t"', "\t"),
    ],
)
def test_token_get_string_literal_value(text: str, expected: float):
    token = get_token(text)
    assert token.kind == MLIRTokenKind.STRING_LIT
    assert token.kind.get_string_literal_value(token.span) == expected


START_END_LINE_CONTENT = """\
0123
5678"""


@pytest.mark.parametrize(
    "pos,expected_start,expected_end",
    [
        (0, 0, 4),
        (1, 0, 4),
        (5, 5, 9),
        (9, 5, 9),
    ],
)
def test_get_start_of_line(
    pos: Position, expected_start: Position, expected_end: Position
):
    input = Input(START_END_LINE_CONTENT, "<>")
    assert input.get_start_of_line(pos) == expected_start
    assert input.get_end_of_line(pos) == expected_end



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass, field
from itertools import repeat
from typing import NamedTuple

from xdsl.builder import ImplicitBuilder
from xdsl.context import Context
from xdsl.dialects import memref_stream
from xdsl.dialects.builtin import (
    AffineMapAttr,
    ArrayAttr,
    IntegerAttr,
    ModuleOp,
)
from xdsl.ir import Block, Region, SSAValue
from xdsl.ir.affine import AffineExpr
from xdsl.passes import ModulePass
from xdsl.pattern_rewriter import (
    PatternRewriter,
    PatternRewriteWalker,
    RewritePattern,
    op_type_rewrite_pattern,
)


def factors(num: int) -> tuple[int, ...]:
    """
    For all positive integers, returns the n-tuple of all numbers that evenly divide the
    input, returns an empty tuple for 0 or negative inputs.
    """
    if num <= 0:
        return ()

    if num == 1:
        return (1,)

    return tuple(factor for factor in range(1, num + 1) if not num % factor)


class IndexAndFactor(NamedTuple):
    """
    Helper data structure holding an option for which index of a `memref_stream.generic`
    operation to interleave and with which factor.
    """

    iterator_index: int
    factor: int

    @staticmethod
    def choose(
        indices_and_factors: Sequence[IndexAndFactor], pipeline_depth: int
    ) -> IndexAndFactor | None:
        """
        A heuristic to choose the interleave index and factor automatically given the
        pipeline depth of floating-point operations on the processor.
        The higher the factor chosen, the higher the instruction-level parallelism, but
        also the more registers need to be used at the same time, potentially leading to
        spilling.
        """
        if not indices_and_factors:
            return None
        # Filter for innermost parallel index
        max_index = max(index for index, _ in indices_and_factors)
        indices_and_factors = tuple(
            t for t in indices_and_factors if t.iterator_index == max_index
        )

        # Want the biggest number for maximal instruction-level parallelism, less than
        # 2 * pipeline depth as a heuristic to limit register pressure.
        indices_and_factors = tuple(
            t for t in indices_and_factors if t.factor < pipeline_depth * 2
        )
        if not indices_and_factors:
            return None

        sorted_indices_and_factors = sorted(indices_and_factors, key=lambda x: x[1])

        # Greatest number less than double of pipeline depth.
        return sorted_indices_and_factors[-1]


@dataclass(frozen=True)
class PipelineGenericPattern(RewritePattern):
    pipeline_depth: int = field()

    @staticmethod
    def indices_and_factors(
        op: memref_stream.GenericOp,
    ) -> tuple[IndexAndFactor, ...]:
        """
        Given a `memref_stream.generic` operation, returns all the possible options for
        unrolling.
        """
        parallel_indices = tuple(
            index
            for index, iterator_type in enumerate(op.iterator_types)
            if iterator_type == memref_stream.IteratorTypeAttr.parallel()
        )
        parallel_bounds = tuple(
            op.bounds.data[index].value.data for index in parallel_indices
        )
        return tuple(
            IndexAndFactor(index, factor)
            for index, bound in zip(parallel_indices, parallel_bounds)
            for factor in factors(bound)
        )

    @op_type_rewrite_pattern
    def match_and_rewrite(
        self, op: memref_stream.GenericOp, rewriter: PatternRewriter
    ) -> None:
        if memref_stream.IteratorTypeAttr.interleaved() in op.iterator_types:
            # Already interleaved
            return

        if memref_stream.IteratorTypeAttr.reduction() not in op.iterator_types:
            # No reduction
            return

        indices_and_factors = self.indices_and_factors(op)
        if not indices_and_factors:
            return

        t = IndexAndFactor.choose(indices_and_factors, self.pipeline_depth)
        if t is None:
            return

        interleave_bound_index, interleave_factor = t

        if interleave_factor == 1:
            # If unroll factor is 1, rewrite is a no-op
            return

        old_block = op.body.block
        new_region = Region(
            Block(
                arg_types=(
                    t
                    for arg in old_block.args
                    for t in repeat(arg.type, interleave_factor)
                )
            )
        )
        with ImplicitBuilder(new_region) as args:
            # For each interleaved block replica, a mapping from old values to new values
            value_map: tuple[dict[SSAValue, SSAValue], ...] = tuple(
                {} for _ in range(interleave_factor)
            )
            for arg_index, new_arg in enumerate(args):
                old_arg = old_block.args[arg_index // interleave_factor]
                value_map[arg_index % interleave_factor][old_arg] = new_arg
                new_arg.name_hint = old_arg.name_hint
            for block_op in old_block.ops:
                if isinstance(block_op, memref_stream.YieldOp):
                    memref_stream.YieldOp(
                        *([vm[arg] for vm in value_map for arg in block_op.arguments])
                    )
                else:
                    for i in range(interleave_factor):
                        block_op.clone(value_mapper=value_map[i])

        # New maps are the same, except that they have one more dimension and the
        # dimension that is interleaved is updated to
        # `dim * interleave_factor + new_dim`.
        new_indexing_maps = ArrayAttr(
            AffineMapAttr(
                m.data.replace_dims_and_symbols(
                    (
                        tuple(
                            AffineExpr.dimension(i)
                            for i in range(interleave_bound_index)
                        )
                        + (
                            AffineExpr.dimension(interleave_bound_index)
                            * interleave_factor
                            + AffineExpr.dimension(m.data.num_dims),
                        )
                        + tuple(
                            AffineExpr.dimension(i)
                            for i in range(
                                interleave_bound_index + 1, m.data.num_dims + 2
                            )
                        )
                    ),
                    (),
                    m.data.num_dims + 1,
                    0,
                )
            )
            for m in op.indexing_maps
        )

        # The new bounds are the same, except there is one more bound
        new_bounds = list(op.bounds)
        new_bounds.append(IntegerAttr.from_index_int_value(interleave_factor))
        interleave_bound = op.bounds.data[interleave_bound_index].value.data
        new_bounds[interleave_bound_index] = IntegerAttr.from_index_int_value(
            interleave_bound // interleave_factor
        )

        rewriter.replace_matched_op(
            memref_stream.GenericOp(
                op.inputs,
                op.outputs,
                op.inits,
                new_region,
                new_indexing_maps,
                ArrayAttr(
                    op.iterator_types.data
                    + (memref_stream.IteratorTypeAttr.interleaved(),)
                ),
                ArrayAttr(new_bounds),
                op.init_indices,
                op.doc,
                op.library_call,
            )
        )


@dataclass(frozen=True)
class MemRefStreamInterleavePass(ModulePass):
    """
    Tiles the innermost parallel dimension of a `memref_stream.generic`.
    If specified, the `pipeline-depth` parameter specifies the number of operations in the
    resulting body that should be executed concurrently.
    The pass will select the largest factor of the corresponding bound smaller than
    `pipeline-depth * 2`.
    The search range is bound by `pipeline-depth * 2` as very large interleaving factors
    can increase register pressure and potentially exhaust all available registers.
    In the future, it would be good to take the number of available registers into account
    when choosing a search range, as well as inspecting the generic body for
    read-after-write dependencies.
    """

    name = "memref-stream-interleave"

    pipeline_depth: int = field(default=4)

    def apply(self, ctx: Context, op: ModuleOp) -> None:
        PatternRewriteWalker(
            PipelineGenericPattern(self.pipeline_depth),
            apply_recursively=False,
        ).rewrite_module(op)



# RUN: python %s | filecheck %s

from xdsl.dialects.builtin import (
    AnyTensorTypeConstr,
    ComplexType,
    Float32Type,
    IndexType,
    IntAttr,
    IntegerAttr,
    IntegerType,
    NoneType,
    Signedness,
)
from xdsl.irdl import (
    AllOf,
    AnyAttr,
    AnyOf,
    AttributeDef,
    AttrSizedOperandSegments,
    BaseAttr,
    EqAttrConstraint,
    OpDef,
    OperandDef,
    OptOperandDef,
    ParamAttrConstraint,
    ParamAttrDef,
    PropertyDef,
    ResultDef,
    SameVariadicOperandSize,
    VarOperandDef,
    traits_def,
)
from xdsl.traits import (
    ConstantLike,
    Pure,
)
from xdsl.utils.dialect_codegen import dump_dialect_pyfile, generate_dynamic_attr_class

types = [
    (
        "Test_SingletonAType",
        ParamAttrDef(name="test.singleton_a", parameters=[]),
    ),
    (
        "Test_SingletonBType",
        ParamAttrDef(name="test.singleton_b", parameters=[]),
    ),
    (
        "Test_SingletonCType",
        ParamAttrDef(name="test.singleton_c", parameters=[]),
    ),
]

SingletonAType = generate_dynamic_attr_class(types[0][0], types[0][1])
SingletonBType = generate_dynamic_attr_class(types[1][0], types[1][1])
SingletonCType = generate_dynamic_attr_class(types[2][0], types[2][1])

attrs = [("Test_TestAttr", ParamAttrDef(name="test.test", parameters=[]))]

ops = [
    (
        "Test_SimpleOp",
        OpDef(name="test.simple", assembly_format="attr-dict"),
    ),
    (
        "Test_AndOp",
        OpDef(
            name="test.and",
            operands=[
                (
                    "in_",
                    OperandDef(
                        AllOf(
                            (
                                AnyAttr(),
                                BaseAttr(SingletonAType),
                            )
                        )
                    ),
                )
            ],
        ),
    ),
    ("Test_AnyOp", OpDef(name="test.any", operands=[("in_", OperandDef(AnyAttr()))])),
    (
        "Test_Integers",
        OpDef(
            name="test.integers",
            operands=[
                (
                    "any_int",
                    OperandDef(
                        ParamAttrConstraint(
                            IntegerType,
                            (EqAttrConstraint(IntAttr(8)), AnyAttr()),
                        )
                    ),
                ),
                ("any_integer", OperandDef(BaseAttr(IntegerType))),
            ],
        ),
    ),
    (
        "Test_OrOp",
        OpDef(
            name="test.or",
            operands=[
                (
                    "in_",
                    OperandDef(
                        AnyOf(
                            (
                                BaseAttr(SingletonAType),
                                BaseAttr(SingletonBType),
                                BaseAttr(SingletonCType),
                            )
                        )
                    ),
                )
            ],
        ),
    ),
    (
        "Test_TypesOp",
        OpDef(
            name="test.types",
            operands=[
                ("a", OperandDef(EqAttrConstraint(IntegerType(32)))),
                ("b", OperandDef(EqAttrConstraint(IntegerType(64, Signedness.SIGNED)))),
                (
                    "c",
                    OperandDef(EqAttrConstraint(IntegerType(8, Signedness.UNSIGNED))),
                ),
                ("d", OperandDef(EqAttrConstraint(IndexType()))),
                ("e", OperandDef(EqAttrConstraint(Float32Type()))),
                ("f", OperandDef(EqAttrConstraint(NoneType()))),
                ("v1", OperandDef(ParamAttrConstraint(ComplexType, (AnyAttr(),)))),
            ],
        ),
    ),
    (
        "Test_SingleOp",
        OpDef(
            name="test.single",
            operands=[("arg", OperandDef(AnyTensorTypeConstr))],
            results=[("res", ResultDef(AnyTensorTypeConstr))],
            assembly_format="$arg attr-dict : type($arg) -> type($res)",
        ),
    ),
    (
        "Test_VariadicityOp",
        OpDef(
            name="test.variadic",
            operands=[
                ("opt", OptOperandDef(BaseAttr(SingletonAType))),
                ("variadic", VarOperandDef(BaseAttr(SingletonAType))),
                ("required", OperandDef(BaseAttr(SingletonCType))),
            ],
            options=[SameVariadicOperandSize(), AttrSizedOperandSegments()],
        ),
    ),
    (
        "Test_PropertiesOp",
        OpDef(
            name="test.properties",
            properties={
                "int_attr": PropertyDef(
                    IntegerAttr.constr(EqAttrConstraint(IntegerType(16)))
                ),
                "in": PropertyDef(AnyAttr()),
            },
            accessor_names={"in_": ("in", "property")},
        ),
    ),
    (
        "Test_TraitsOp",
        OpDef(
            name="test.traits",
            traits=traits_def(ConstantLike(), Pure()),
        ),
    ),
    (
        "Test_AttributesOp",
        OpDef(
            name="test.attributes",
            attributes={
                "attr": AttributeDef(AnyAttr()),
                "other_attr": AttributeDef(AnyAttr()),
            },
            accessor_names={"some_attr": ("attr", "attribute")},
        ),
    ),
]

dump_dialect_pyfile(
    "test",
    ops,
    attributes=attrs,
    types=types,
)

# CHECK:       from xdsl.dialects.builtin import *
# CHECK-NEXT:  from xdsl.ir import *
# CHECK-NEXT:  from xdsl.irdl import *
# CHECK-NEXT:  from xdsl.traits import *

# CHECK:       # ruff: noqa: F403, F405

# CHECK:       @irdl_attr_definition
# CHECK-NEXT:  class Test_SingletonAType(ParametrizedAttribute, TypeAttribute):
# CHECK-NEXT:      name = "test.singleton_a"

# CHECK:       @irdl_attr_definition
# CHECK-NEXT:  class Test_SingletonBType(ParametrizedAttribute, TypeAttribute):
# CHECK-NEXT:      name = "test.singleton_b"

# CHECK:       @irdl_attr_definition
# CHECK-NEXT:  class Test_SingletonCType(ParametrizedAttribute, TypeAttribute):
# CHECK-NEXT:      name = "test.singleton_c"

# CHECK:       @irdl_attr_definition
# CHECK-NEXT:  class Test_TestAttr(ParametrizedAttribute):
# CHECK-NEXT:      name = "test.test"

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_SimpleOp(IRDLOperation):
# CHECK-NEXT:      name = "test.simple"

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_AndOp(IRDLOperation):
# CHECK-NEXT:      name = "test.and"
# CHECK-NEXT:      in_ = operand_def(AllOf(attr_constrs=(AnyAttr(), BaseAttr(Test_SingletonAType))))

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_AnyOp(IRDLOperation):
# CHECK-NEXT:      name = "test.any"
# CHECK-NEXT:      in_ = operand_def(AnyAttr())

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_Integers(IRDLOperation):
# CHECK-NEXT:      name = "test.integers"
# CHECK-NEXT:      any_int = operand_def(
# CHECK-NEXT:          ParamAttrConstraint(
# CHECK-NEXT:              IntegerType, (EqAttrConstraint(attr=IntAttr(data=8)), AnyAttr())
# CHECK-NEXT:          )
# CHECK-NEXT:      )
# CHECK-NEXT:      any_integer = operand_def(BaseAttr(IntegerType))

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_OrOp(IRDLOperation):
# CHECK-NEXT:      name = "test.or"
# CHECK-NEXT:      in_ = operand_def(
# CHECK-NEXT:          AnyOf(
# CHECK-NEXT:              attr_constrs=(
# CHECK-NEXT:                  BaseAttr(Test_SingletonAType),
# CHECK-NEXT:                  BaseAttr(Test_SingletonBType),
# CHECK-NEXT:                  BaseAttr(Test_SingletonCType),
# CHECK-NEXT:              )
# CHECK-NEXT:          )
# CHECK-NEXT:      )

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_TypesOp(IRDLOperation):
# CHECK-NEXT:      name = "test.types"
# CHECK-NEXT:      a = operand_def(EqAttrConstraint(attr=IntegerType(32)))
# CHECK-NEXT:      b = operand_def(EqAttrConstraint(attr=IntegerType(64, Signedness.SIGNED)))
# CHECK-NEXT:      c = operand_def(EqAttrConstraint(attr=IntegerType(8, Signedness.UNSIGNED)))
# CHECK-NEXT:      d = operand_def(EqAttrConstraint(attr=IndexType()))
# CHECK-NEXT:      e = operand_def(EqAttrConstraint(attr=Float32Type()))
# CHECK-NEXT:      f = operand_def(EqAttrConstraint(attr=NoneType()))
# CHECK-NEXT:      v1 = operand_def(ParamAttrConstraint(ComplexType, (AnyAttr(),)))

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_SingleOp(IRDLOperation):
# CHECK-NEXT:      name = "test.single"
# CHECK-NEXT:      arg = operand_def(BaseAttr(TensorType))
# CHECK-NEXT:      res = result_def(BaseAttr(TensorType))
# CHECK-EMPTY:
# CHECK-NEXT:      assembly_format = "$arg attr-dict : type($arg) -> type($res)"


# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_VariadicityOp(IRDLOperation):
# CHECK-NEXT:      name = "test.variadic"
# CHECK-NEXT:      opt = opt_operand_def(BaseAttr(Test_SingletonAType))
# CHECK-NEXT:      variadic = var_operand_def(BaseAttr(Test_SingletonAType))
# CHECK-NEXT:      required = operand_def(BaseAttr(Test_SingletonCType))
# CHECK-EMPTY:
# CHECK-NEXT:      irdl_options = [
# CHECK-NEXT:          SameVariadicOperandSize(),
# CHECK-NEXT:          AttrSizedOperandSegments(as_property=False),
# CHECK-NEXT:      ]

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_PropertiesOp(IRDLOperation):
# CHECK-NEXT:      name = "test.properties"
# CHECK-EMPTY:
# CHECK-NEXT:      int_attr = prop_def(
# CHECK-NEXT:          ParamAttrConstraint(
# CHECK-NEXT:              IntegerAttr, (AnyAttr(), EqAttrConstraint(attr=IntegerType(16)))
# CHECK-NEXT:          )
# CHECK-NEXT:      )
# CHECK-NEXT:      in_ = prop_def(AnyAttr(), prop_name="in")

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_TraitsOp(IRDLOperation):
# CHECK-NEXT:      name = "test.traits"
# CHECK-EMPTY:
# CHECK-NEXT:      traits = traits_def(ConstantLike(), Pure())

# CHECK:       @irdl_op_definition
# CHECK-NEXT:  class Test_AttributesOp(IRDLOperation):
# CHECK-NEXT:      name = "test.attributes"
# CHECK-EMPTY:
# CHECK-NEXT:      some_attr = attr_def(AnyAttr(), attr_name="attr")
# CHECK-NEXT:      other_attr = attr_def(AnyAttr())

# CHECK:       TestDialect = Dialect(
# CHECK-NEXT:      "test",
# CHECK-NEXT:      [
# CHECK-NEXT:          Test_SimpleOp,
# CHECK-NEXT:          Test_AndOp,
# CHECK-NEXT:          Test_AnyOp,
# CHECK-NEXT:          Test_Integers,
# CHECK-NEXT:          Test_OrOp,
# CHECK-NEXT:          Test_TypesOp,
# CHECK-NEXT:          Test_SingleOp,
# CHECK-NEXT:          Test_VariadicityOp,
# CHECK-NEXT:          Test_PropertiesOp,
# CHECK-NEXT:          Test_TraitsOp,
# CHECK-NEXT:          Test_AttributesOp,
# CHECK-NEXT:      ],
# CHECK-NEXT:      [Test_TestAttr, Test_SingletonAType, Test_SingletonBType, Test_SingletonCType],
# CHECK-NEXT:  )



# RUN: python %s | filecheck %s
from pathlib import Path

from xdsl.context import Context
from xdsl.dialects import get_all_dialects
from xdsl.dialects.irdl.irdl import DialectOp
from xdsl.interpreters.irdl import make_dialect
from xdsl.parser import Parser
from xdsl.traits import SymbolTable
from xdsl.utils.dialect_stub import DialectStubGenerator

# Test dynamically generating Python type stubs for an irdl dialect.

if __name__ == "__main__":
    # Register all dialects for lazy-loading
    ctx = Context()
    for n, f in get_all_dialects().items():
        # Except cmath to avoid conflict with the one we're going to load from its IRDL description
        if n == "cmath":
            continue
        ctx.register_dialect(n, f)

    # Open the IRDL description of cmath, parse it
    f = (Path(__file__).parent / "cmath.irdl.mlir").open()
    parser = Parser(ctx, f.read())
    module = parser.parse_module()

    # Make it a PyRDL Dialect
    dialect_op = SymbolTable.lookup_symbol(module, "cmath")
    assert isinstance(dialect_op, DialectOp)
    dialect = make_dialect(dialect_op)

    # Generate and print type stubs!
    stub = DialectStubGenerator(dialect)
    print(stub.generate_dialect_stubs())

# CHECK:       from xdsl.dialects.builtin import (
# CHECK-NEXT:      Float32Type,
# CHECK-NEXT:      Float64Type,
# CHECK-NEXT:  )
# CHECK-NEXT:  from xdsl.ir import (
# CHECK-NEXT:      Dialect,
# CHECK-NEXT:      OpResult,
# CHECK-NEXT:      ParametrizedAttribute,
# CHECK-NEXT:      TypeAttribute,
# CHECK-NEXT:  )
# CHECK-NEXT:  from xdsl.irdl import (
# CHECK-NEXT:      IRDLOperation,
# CHECK-NEXT:      Operand,
# CHECK-NEXT:  )
# CHECK-EMPTY:
# CHECK-NEXT:  class complex(TypeAttribute, ParametrizedAttribute):
# CHECK-NEXT:      elem : "Float32Type | Float64Type"
# CHECK-EMPTY:
# CHECK-EMPTY:
# CHECK-NEXT:  class NormOp(IRDLOperation):
# CHECK-NEXT:      in_ : Operand
# CHECK-NEXT:      out : OpResult
# CHECK-EMPTY:
# CHECK-EMPTY:
# CHECK-NEXT:  class MulOp(IRDLOperation):
# CHECK-NEXT:      lhs : Operand
# CHECK-NEXT:      rhs : Operand
# CHECK-NEXT:      res : OpResult
# CHECK-EMPTY:
# CHECK-EMPTY:
# CHECK-NEXT:  Cmath : Dialect
# CHECK-EMPTY:



# TID 251 enforces to not import from those
# We need to skip it here to allow importing from here instead.
from .dynamic_index_list import *  # noqa: TID251
from .fast_math import *  # noqa: TID251
from .format import *  # noqa: TID251
